/++
 + FastCGIフレームワーク Raptan Version: 0.3.3
 +
 + 基本機能モジュール
 +
 + Copyright(C) 2009-2011 Takuya Kurosawa
 +/
module jp.ku6.raptan.core;

import core.thread;
import core.stdc.stdlib;
import std.conv;
import std.datetime;
import std.string;
import std.stream;
import std.intrinsic;
import com.fastcgi.fcgiapp;

/++
 + クライアントからのリクエスト時の解析エラー
 +/
class RequestParseError : Error
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  message = メッセージ
	 +/
	this(string message)
	{
		super(message);
	}
}

/++
 + アップロードされたファイル
 +
 + 小さいファイルはメモリ上、大きいファイルはディスク上と切り替えてファイルを保持する
 +/
class UploadedFile
{
	/++
	 + 名称を取得
	 + 
	 +/
	const
	string name()
	nothrow
	{
		return name_;
	}

	/++
	 + ファイルがメモリ上にあるか判定
	 +
	 + メモリ上にあればdataで実体を
	 + ディスク上にあればfilePathでファイルパスを取得する
	 +
	 + Returns:
	 +  ファイルがメモリ上にあればtrueを返す
	 +/
	const
	bool onMemory()
	nothrow
	{
		return data_ !is null;
	}

	/++
	 + アップロードされたファイルのサイズを取得
	 +
	 + Returns:
	 +  アップロードされたファイルのサイズ
	 +/
	const
	uint size()
	nothrow
	{
		return writeCursor_;
	}

	/++
	 + ファイルの実体データを取得
	 +
	 + Returns:
	 +  ファイルの実体データ
	 +/
	ubyte[] data()
	nothrow
	{
		return data_[0 .. writeCursor_];
	}

	/++
	 + ファイルのパスを取得
	 +
	 + Returns:
	 +  ファイルパス
	 +/
	const
	string filePath()
	nothrow
	{
		return filePath_;
	}

private:

	// アップロードされたファイル名
	string name_;

	// メモリ上に保持する最大サイズ
	uint threshold_;

	// 出力ファイルのディレクトリ
	string directory_;

	// 書き込んだデータサイズ
	uint writeCursor_;

	// 実体データ
	ubyte* data_;

	// 実態データのサイズ
	uint dataSize_;

	// ファイルパス
	string filePath_;

	// 書き込みファイル
	File file_;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  name = 名称
	 +  threshod = メモリ上に保持する最大データサイズ
	 +  directory = ディスク上に一時ファイルを書き出す時のディレクトリ
	 +/
	this(string name, uint threshold, string directory)
	{
		name_ = name;
		threshold_ = threshold;
		directory_ = directory;
	}

	/+
	 + ディスクに書き出す時のファイルパスを生成
	 +
	 + 他のファイルと重複しないパスを生成する
	 +
	 + Returns:
	 +  ファイルパス
	 +/
	string generateUniqueFilePath()
	{
		// 100%じゃないけど現実的に大丈夫だろうから今はこれでいい。
		return std.path.buildPath(directory_, std.string.format("uploaded_%016x.bin", Clock.currTime().stdTime));
	}

	/+
	 + データを追加する
	 +
	 + Params:
	 +  source = 追加するデータ
	 +/
	void append(const(ubyte)[] source)
	{
		if (source is null || source.length == 0)
			return;

		if (writeCursor_ > threshold_)
		{
			// ファイルに追記
			uint written;
			while (written < source.length)
				written += file_.write(source[written .. $]);
			writeCursor_ += source.length;
		}
		else
		{
			if (writeCursor_ + source.length > threshold_)
			{
				filePath_ = generateUniqueFilePath();
				file_ = new File(filePath_, FileMode.OutNew);

				// 既にバッファ上にあるデータをファイルに書き込み
				if (data_ !is null)
				{
					uint written;
					while (written < writeCursor_)
						written += file_.write(data_[written .. writeCursor_]);
					core.stdc.stdlib.free(data_);
					data_ = null;
				}

				// 今回のデータをファイルに追記
				uint written;
				while (written < source.length)
					written += file_.write(source[written .. $]);
				writeCursor_ += written;
			}
			else
			{
				// 今回のデータを足してもメモリに入りきるので、メモリを伸ばす

				// メモリの確保
				if (data_ is null)
				{
					data_ = cast(ubyte*)core.stdc.stdlib.malloc(source.length);
				}
				else
				{
					uint newDataSize = dataSize_;
					while (writeCursor_ + source.length > newDataSize)
					{
						newDataSize = newDataSize * 2 + 1;
						if (newDataSize > threshold_)
							newDataSize = threshold_;
					}

					if (newDataSize != dataSize_)
					{
						data_ = cast(ubyte*)core.stdc.stdlib.realloc(data_, newDataSize);
						dataSize_ = newDataSize;
					}
				}

				// データをコピー
				ubyte* destination = &data_[writeCursor_];
				size_t length = source.length;
				for (size_t i = 0; i < length; ++i)
					destination[i] = source[i];
				writeCursor_ += length;
			}
		}
	}

	/+
	 + ファイルが開かれていたら閉じる
	 +/
	void closeFileIfOpen()
	{
		if (file_ !is null)
		{
			file_.close();
			file_ = null;
		}
	}

	/+
	 + 確保したメモリ・ファイルを削除する
	 +/
	void cleanup()
	{
		if (filePath_ !is null && std.file.exists(filePath_))
		{
			std.file.remove(filePath_);
			filePath_ = null;
		}

		if (data_ !is null)
		{
			core.stdc.stdlib.free(data_);
		}
	}
}

/++
 + リクエストされた情報を保持するクラス
 +/
class Request
{
	/++
	 + リクエストメソッド
	 +/
	enum Method
	{
		UNKNOWN = 1,
		GET = 2,
		POST = 4,
		PUT = 8,
		HEAD = 16,
		DELETE = 32,
		TRACE = 64,
		OPTIONS = 128,
		CONNECT = 256,
	}

	/++
	 + リクエストメソッドを取得
	 +
	 + Returns:
	 +  リクエストメソッド
	 +/
	const
	Method method()
	nothrow
	{
		return method_;
	}

	/++
	 + リクエストURIを取得
	 +
	 + Returns:
	 +  リクエストURI
	 +/
	const
	string uri()
	nothrow
	{
		return uri_;
	}

	/++
	 + リクエストパラメータから指定キーの値を1つ取得
	 +
	 + 複数ある場合は先頭のものを返す
	 +
	 + Params:
	 +  key = key
	 +
	 + Returns:
	 +  対応する値 存在しない場合はnullを返す
	 +/
	const
	string getParameter(string key)
	{
		const(const(char)[][])* values = key in parameters_;
		if (values is null)
			return null;
		else
			return (*values)[0].idup;
	}

	/++
	 + リクエストパラメータから指定キーの値を1つ取得(内部バッファ参照版)
	 +
	 + 内部バッファを直接参照するため、getParameterに比べてバッファの作成&コピーコストが削減出来る
	 + 内部バッファは次のリクエストを受け取るときに上書きされるため、
	 + 返される値は1リクエストの間だけ使用すること
	 +
	 + 複数ある場合は先頭のものを返す
	 +
	 + Params:
	 +  key = key
	 +
	 + Returns:
	 +  対応する値 存在しない場合はnullを返す
	 +/
	const
	const(char)[] getParameterDirect(string key)
	nothrow
	{
		const(const(char)[][])* values = key in parameters_;
		if (values is null)
			return null;
		else
			return (*values)[0];
	}

	/++
	 + リクエストパラメータから指定キーの値をT型として取得
	 +
	 + 複数ある場合は先頭のものを返す
	 +
	 + Params:
	 +  key = key
	 +  defaultValue = キーの値が無い場合や、変換出来ない場合に返す値
	 +
	 + Returns:
	 +  対応する値 存在しない場合や変換出来ない場合はdefaultValueを返す
	 +/
	const
	T getParameterAs(T)(string key, T defaultValue=T.init)
	{
		const(const(char)[][])* value = key in parameters_;
		if (value is null)
			return defaultValue;

		try
			return to!(T)((*value)[0]);
		catch (Error e)
			return defaultValue;
	}

	/+
	 + getParameterAsのbool型への変換の特殊化
	 +
	 + trueと判定する: 1, t, true
	 + falseと判定する: 0, f, false
	 + ※大文字小文字区別なし
	 +
	 + Params:
	 +  key = パラメータのキー
	 +  defaultValue = キーの値が無い場合や、変換出来ない場合に返す値
	 +/
	const
	T getParameterAs(T:bool)(string key, T defaultValue=T.init)
	{
		const(const(char)[][])* value = key in parameters_;
		if (value is null)
			return defaultValue;

		const(char)[] v = (*value)[0];
		if (cmp(v, "1") == 0 || icmp(v, "t") == 0 || icmp(v, "true") == 0)
			return true;
		else if (cmp(v, "0") == 0 || icmp(v, "f") == 0 || icmp(v, "false") == 0)
			return false;
		else
			return defaultValue;
	}

	/++
	 + リクエストパラメータから指定キーの値を全て取得
	 +
	 + Params:
	 +  key = key
	 +
	 + Returns:
	 +  対応する値 存在しない場合は空の配列を返す
	 +/
	const
	string[] getParameters(string key)
	{
		const(const(char)[][])* values = key in parameters_;
		if (values is null)
			return [];
		else
		{
			size_t l = (*values).length;
			string[] result = new string[l];
			for (size_t i; i < l; ++i)
				result[i] = (*values)[i].idup;
			return result;
		}
	}

	/++
	 + リクエストパラメータから指定キーの値を全て取得(内部バッファ参照版)
	 +
	 + 内部バッファを直接参照するため、getParametersに比べてバッファの作成&コピーコストが削減出来る
	 + 内部バッファは次のリクエストを受け取るときに上書きされるため、
	 + 返される値は1リクエストの間だけ使用すること
	 +
	 + Params:
	 +  key = key
	 +
	 + Returns:
	 +  対応する値 存在しない場合は空の配列を返す
	 +/
	const
	const(const(char)[][]) getParametersDirect(string key)
	nothrow
	{
		const(const(char)[][])* values = key in parameters_;
		if (values is null)
			return [];
		else
			return (*values);
	}

	/++
	 + リクエストパラメータから指定キーの値をT型として全て取得
	 +
	 + Params:
	 +  key = key
	 +  defaultValue = 変換出来ない場合に返す値
	 +
	 + Returns:
	 +  対応する値 存在しない場合は空の配列を返す
	 +/
	const
	T[] getParametersAs(T)(string key, T defaultValue=T.init)
	{
		T[] result = [];

		const(const(char)[][])* values = key in parameters_;
		if (values is null)
			return result;

		foreach (v; *values)
		{
			try
				result ~= to!(T)(v);
			catch (Error e)
				result ~= defaultValue;
		}

		return result;
	}

	/+
	 + getParametersAsのbool型への変換の特殊化
	 +
	 + trueと判定する: 1, t, true
	 + falseと判定する: 0, f, false
	 + ※大文字小文字区別なし
	 +
	 + Params:
	 +  key = パラメータのキー
	 +  defaultValue = 変換出来ない場合に返す値
	 +
	 + Returns:
	 +  対応する値 存在しない場合は空の配列を返す
	 +/
	const
	T[] getParametersAs(T:bool)(string key, T defaultValue=T.init)
	{
		T[] result = [];

		const(const(char)[][])* values = key in parameters_;
		if (values is null)
			return result;

		foreach (v; *values)
		{
			if (cmp(v, "1") == 0 || icmp(v, "t") == 0 || icmp(v, "true") == 0)
				result ~= true;
			else if (cmp(v, "0") == 0 || icmp(v, "f") == 0 || icmp(v, "false") == 0)
				result ~= false;
			else
				result ~= defaultValue;
		}

		return result;
	}

	/++
	 + Cookieから指定キーの値を全て取得
	 +
	 + Params:
	 +  key = key
	 +
	 + Returns:
	 +  対応する値の配列 存在しない場合は空の配列を返す
	 +/
	const
	const(string[]) getCookieValues(string key)
	{
		const(const(char)[][])* values = key in cookieValues_;
		if (values is null)
			return [];
		else
		{
			size_t l = (*values).length;
			string[] result = new string[l];
			for (size_t i; i < l; ++i)
				result[i] = (*values)[i].idup;
			return result;
		}
	}

	/++
	 + Cookieから指定キーの値を全て取得
	 +
	 + 内部バッファを直接参照するため、getCookieValuesに比べてバッファの作成&コピーコストが削減出来る
	 + 内部バッファは次のリクエストを受け取るときに上書きされるため、
	 + 返される値は1リクエストの間だけ使用すること
	 +
	 + Params:
	 +  key = key
	 +
	 + Returns:
	 +  対応する値の配列 存在しない場合は空の配列を返す
	 +/
	const
	const(const(char)[][]) getCookieValuesDirect(string key)
	nothrow
	{
		const(const(char)[][])* values = key in cookieValues_;
		if (values is null)
			return [];
		else
			return (*values);
	}

	/++
	 + アップロードされたファイルを取得
	 +
	 + Params:
	 +  key = key
	 +
	 + Returns:
	 +  アップロードされたファイル 無い場合はnull
	 +/
	UploadedFile getUploadedFile(string key)
	{
		UploadedFile* file = key in uploadedFiles_;
		if (file is null)
			return null;

		return *file;
	}

	/++
	 + 環境変数を取得
	 +
	 + Params:
	 +  key = キー文字列
	 +
	 + Returns:
	 +  値
	 +/
	string getEnvironmentVariable(string key)
	{
		return std.conv.to!string(FCGX_GetParam(std.string.toStringz(key), fcgxRequest_.envp));
	}

	/++
	 + リクエストURIを書き換える
	 +
	 + 書き換えないほうがいいに決まってるけど
	 + どうしても書き換えたいときもある。
	 +
	 + Params:
	 +  uri = リクエストURI
	 +/
	void setUri(string uri)
	{
		uri_ = uri;
	}

	/++
	 + リクエストパラメータを書き換える
	 +
	 + 書き換えないほうがいいに決まってるけど
	 + どうしても書き換えたいときもある。
	 +
	 + Params:
	 +  key = key
	 +  value = value
	 +/
	void setParameter(string key, string value)
	{
		if ((key in parameters_) is null)
			parameters_[key] = [value];
		else
			parameters_[key] ~= value;
	}

private:

	// FCGXのRequest構造体
	FCGX_Request fcgxRequest_;

	// リクエストメソッド
	Method method_;

	// リクエストURI
	string uri_;

	// パラメータ
	const(char)[][][const(char)[]] parameters_;

	// Cookieの値
	const(char)[][][const(char)[]] cookieValues_;

	// アップロードされたファイル
	UploadedFile[string] uploadedFiles_;

	// multipart/form-data解析用クラス
	MultipartFormDataCursor cursor_;

	// POSTで渡されたデータを解析するためのバッファ
	char* buffer_;

	// バッファサイズ
	uint bufferSize_;

	/+
	 + コンストラクタ
	 +
	 + Params:
	 +  bufferSize = POSTで渡されたデータを解析するためのバッファサイズ
	 +  uploadFileOnMemoryThreshold = アップロードされたファイルをメモリに乗せる時の最大値 0を指定すると常にファイルに保存する
	 +  uploadDirectory = アップロードされたファイルを保存するディレクトリ
	 +/
	this(uint bufferSize, uint uploadFileOnMemoryThreshold, string uploadDirectory)
	{
		// バッファは2の累乗に設定
		uint top = std.intrinsic.bsr(bufferSize);
		bufferSize += (1 << top) - 1;
		bufferSize &= 0xffffffff << top;

		bufferSize_ = bufferSize;
		buffer_ = cast(char*)core.stdc.stdlib.malloc(bufferSize_);

		cursor_ = new MultipartFormDataCursor((cast(ubyte*)buffer_)[0 .. bufferSize_], uploadFileOnMemoryThreshold, uploadDirectory);

		FCGX_InitRequest(&fcgxRequest_, 0, 0);
	}

	/+
	 + デストラクタ
	 +/
	~this()
	{
		core.stdc.stdlib.free(buffer_);
	}

	/+
	 + 次のリクエストを取得
	 +/
	bool accept()
	{
		// リクエストパラメータをクリア
		parameters_ = null;

		// 前にアップロードされたファイルをクリアしておく
		if (uploadedFiles_ !is null)
		{
			foreach (file; uploadedFiles_)
				file.cleanup();
			uploadedFiles_ = null;
		}

		// 次のリクエストを待つ
		int result = FCGX_Accept_r(&fcgxRequest_);

		// リクエストメソッドを取得
		char* method = FCGX_GetParam("REQUEST_METHOD\0".ptr, fcgxRequest_.envp);
		if (method[0] == 'G' && method[1] == 'E' && method[2] == 'T' && method[3] == 0)
			method_ = Method.GET;
		else if (method[0] == 'P' && method[1] == 'O' && method[2] == 'S' && method[3] == 'T' && method[4] == 0)
			method_ = Method.POST;
		else if (method[0] == 'P' && method[1] == 'U' && method[2] == 'T' && method[3] == 0)
			method_ = Method.PUT;
		else if (method[0] == 'H' && method[1] == 'E' && method[2] == 'A' && method[3] == 'D' && method[4] == 0)
			method_ = Method.HEAD;
		else if (method[0] == 'D' && method[1] == 'E' && method[2] == 'L' && method[3] == 'E' && method[4] == 'T' && method[5] == 'E' && method[6] == 0)
			method_ = Method.DELETE;
		else if (method[0] == 'T' && method[1] == 'R' && method[2] == 'A' && method[3] == 'C' && method[4] == 'E' && method[5] == 0)
			method_ = Method.TRACE;
		else if (method[0] == 'O' && method[1] == 'P' && method[2] == 'T' && method[3] == 'I' && method[4] == 'O' && method[5] == 'N' && method[6] == 'S' && method[7] == 0)
			method_ = Method.OPTIONS;
		else if (method[0] == 'C' && method[1] == 'O' && method[2] == 'N' && method[3] == 'N' && method[4] == 'E' && method[5] == 'C' && method[6] == 'T' && method[7] == 0)
			method_ = Method.CONNECT;
		else
			method_ = Method.UNKNOWN;

		// リクエストURIを取得
		uri_ = getEnvironmentVariable("SCRIPT_NAME");

		// リクエストパラメータを取得
		if (method_ == Method.GET)
		{
			// QUERY_STRINGを取得
			char* queryString = FCGX_GetParam("QUERY_STRING\0".ptr, fcgxRequest_.envp);

			// リクエストパラメータを解析する
			parameters_ = parseQueryString(queryString);
		}
		else if (method_ == Method.POST)
		{
			// Content-Typeを取得
			string contentType = std.conv.to!string(FCGX_GetParam("CONTENT_TYPE\0".ptr, fcgxRequest_.envp));
			if (std.string.indexOf(contentType, "multipart/form-data;") >= 0)
			{
				// カーソルにストリームをセットして初期化
				cursor_.initialize(fcgxRequest_.input);

				// boundaryを取得
				string boundary = "boundary=";
				ptrdiff_t position = std.string.indexOf(contentType, boundary);
				if (position > 0)
				{
					position += boundary.length;

					// boundaryを指定して解析
					try
					{
						// データを解析
						auto pair = parseMultipartFormData(contentType[position .. $], cursor_);

						// 結果を保存
						parameters_ = pair.parameters;
						uploadedFiles_ = pair.uploadedFiles;
					}
					catch (Throwable t)
					{
						// 特に何もしない
					}
				}

				if (parameters_ is null)
					parameters_[null] = null;

				if (uploadedFiles_ is null)
					uploadedFiles_[null] = null;
			}
			else
			{
				// まずCONTENT_LENGTHを取る
				string stringContentLength = std.conv.to!string(FCGX_GetParam("CONTENT_LENGTH\0".ptr, fcgxRequest_.envp));
				int contentLength;
				try
					contentLength = std.conv.to!(int)(stringContentLength);
				catch (Error e)
					contentLength = 0;

				if (contentLength <= 0)
					parameters_[null] = null;
				else if (contentLength + 1 > bufferSize_)
					parameters_[null] = null;
				else
				{
					// 入力用領域を作ってデータをコピーする
					contentLength = FCGX_GetStr(buffer_, contentLength, fcgxRequest_.input);

					// 最後に0を追加しておく
					buffer_[contentLength] = 0;

					// リクエストパラメータを解析する
					parameters_ = parseQueryString(buffer_);
				}
			}
		}

		// HTTP_COOKIEを取得
		char* cookieString = FCGX_GetParam("HTTP_COOKIE\0".ptr, fcgxRequest_.envp);

		// Cookieを解析する (コピーの速度をケチってバッファを破壊する仕様です。不都合が出たら直します)
		cookieValues_ = parseCookieString(cookieString);

		return result >= 0;
	}

	/+
	 + Query Stringをキーと値の連想配列にパースする
	 +
	 + 引数のバッファをワークエリアとして破壊する
	 + 予めワークエリアにコピーしてから実行すること
	 + 元のバッファサイズを超えてアクセスするようなことはない
	 +
	 + Params:
	 +  queryString = 解析対象の文字列 (Cの文字列)
	 +
	 + Returns:
	 +  個々のパラメータに分けられたデータの連想配列
	 +/
	static
	const(char)[][][const(char)[]] parseQueryString(char* queryString)
	{
		// 解析結果のデータを格納する
		const(char)[][][const(char)[]] parameter;

		// 入力がなければ空を返す
		if (queryString is null)
		{
			parameter[null] = null;
			return parameter;
		}

		char[4] multiByteCharactor;
		uint multiByteCharactorLength;

		bool readKey = true;
		const(char)[] key;

		uint begin;
		uint write;
		for (uint read; ; ++read)
		{
			char c = queryString[read];

			if (c == 0)
				break;

			if (charactorIsControl(c))
				throw new RequestParseError("Illegal charactor found in query string." ~ c);

			// 先頭の?は無視する
			if (c == '?' && write == 0 && queryString[read+1] != 0)
				continue;

			if (c == '%')
			{
				// %が来て、後ろの2文字で16進数が構成できるならばデコードする
				// 後ろ2文字で16進数が構成できない場合は何もしない

				// 0-9か判定
				ubyte upper = cast(ubyte)(queryString[read+1] - '0');
				if (upper < 10)
					upper *= 16;
				else
				{
					// A-Fか判定
					upper -= cast(ubyte)'A' - cast(ubyte)'0';
					if (upper < 6)
					{
						upper += 10;
						upper <<= 4;
					}
					else
					{
						// a-fか判定
						upper -= cast(ubyte)'a' - cast(ubyte)'A';
						if (upper < 6)
						{
							upper += 10;
							upper <<= 4;
						}
						else
						{
							// 16進数でないのでエンコードされていないと判定
							queryString[write++] = c;
							multiByteCharactorLength = 0;
							continue;
						}
					}
				}

				// 0-9か判定
				ubyte lower = cast(ubyte)(queryString[read+2] - '0');
				if (lower < 10)
					upper += lower;
				else
				{
					// A-Fか判定
					lower -= cast(ubyte)'A' - cast(ubyte)'0';
					if (lower < 6)
						upper += lower + 10;
					else
					{
						// a-fか判定
						lower -= cast(ubyte)'a' - cast(ubyte)'A';
						if (lower < 6)
							upper += lower + 10;
						else
						{
							// 16進数でないのでエンコードされていないと判定
							queryString[write++] = c;
							multiByteCharactorLength = 0;
							continue;
						}
					}
				}

				// マルチバイト文字か判定
				if (upper >= 0x80)
				{
					if (multiByteCharactorLength == 0)
					{
						// UTF-8マルチバイト文字1文字目として正当か判定
						if ((upper & 0b11100000) == 0b11000000 || (upper & 0b11110000) == 0b11100000 || (upper & 0b11111000) == 0b11110000)
						{
							multiByteCharactor[multiByteCharactorLength++] = cast(char)upper;
						}
					}
					else
					{
						// UTF-8マルチバイト文字2文字目以降として正当か判定
						if ((upper & 0b11000000) == 0b10000000)
						{
							multiByteCharactor[multiByteCharactorLength++] = cast(char)upper;
							if (multiByteCharactorLength == 2 && (multiByteCharactor[0] & 0b11100000) == 0b11000000)
							{
								queryString[write++] = multiByteCharactor[0];
								queryString[write++] = multiByteCharactor[1];

								multiByteCharactorLength = 0;
							}
							else if (multiByteCharactorLength == 3 && (multiByteCharactor[0] & 0b11110000) == 0b11100000)
							{
								queryString[write++] = multiByteCharactor[0];
								queryString[write++] = multiByteCharactor[1];
								queryString[write++] = multiByteCharactor[2];

								multiByteCharactorLength = 0;
							}
							else if (multiByteCharactorLength == 4)
							{
								queryString[write++] = multiByteCharactor[0];
								queryString[write++] = multiByteCharactor[1];
								queryString[write++] = multiByteCharactor[2];
								queryString[write++] = multiByteCharactor[3];

								multiByteCharactorLength = 0;
							}
						}
						else
						{
							multiByteCharactorLength = 0;
						}
					}
				}
				else
				{
					// 途中で切れたら不正なので無かったことにする
					multiByteCharactorLength = 0;

					// 不正な文字でなかったら書き込む
					if (cast(ubyte)((upper ^ 0x40) - 0x3f) > 0x20 || upper == 0x09 || upper == 0x0a || upper == 0x0d)
						queryString[write++] = cast(char)upper;
				}

				read += 2;

				continue;
			}

			queryString[write] = c;
			multiByteCharactorLength = 0;

			if (c == '=')
			{
				// 値読み込み時に'='が来たら値の一部とする
				if (readKey)
				{
					if (begin == write)
						key = "";
					else
						key = cast(const(char)[])queryString[begin .. write];
					readKey = false;
					begin = write+1;
				}
			}
			else if (c == '&')
			{
				// キー読み込み時に'&'が来たら値は空文字列とする
				if (readKey)
				{
					if (begin < write)
					{
						const(char)[] k = cast(const(char)[])queryString[begin .. write];
						if ((k in parameter) is null)
							parameter[k] = [""];
						else
							parameter[k] ~= "";
					}
				}
				// 値読み込み時に'&'が来たら値はそこまでとする
				else
				{
					if (begin == write)
					{
						if ((key in parameter) is null)
							parameter[key] = [""];
						else
							parameter[key] ~= "";
					}
					else
					{
						if ((key in parameter) is null)
							parameter[key] = [cast(const(char)[])queryString[begin .. write]];
						else
							parameter[key] ~= cast(const(char)[])queryString[begin .. write];
					}

					readKey = true;
				}

				begin = write+1;
			}
			else if (c == '+')
			{
				// '+'は半角スペースに変換する
				queryString[write] = ' ';
			}

			++write;
		}

		// 最後に読み込み中であったものを処理する
		if (readKey)
		{
			if (begin < write)
			{
				const(char)[] k = cast(const(char)[])queryString[begin .. write];
				if ((k in parameter) is null)
					parameter[k] = [""];
				else
					parameter[k] ~= "";
			}
		}
		else
		{
			if (begin == write)
			{
				if ((key in parameter) is null)
					parameter[key] = [""];
				else
					parameter[key] ~= "";
			}
			else
			{
				if ((key in parameter) is null)
					parameter[key] = [cast(const(char)[])queryString[begin .. write]];
				else
					parameter[key] ~= cast(const(char)[])queryString[begin .. write];
			}
		}

		return parameter;
	}
	unittest
	{
		// パラメータのパースをテスト
		const(char)[][][const(char)[]]	parameter;

		parameter = parseQueryString("aaa=bbb\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);

		parameter = parseQueryString("?aaa=bbb\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);

		parameter = parseQueryString("??aaa=bbb&\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);

		parameter = parseQueryString("?aaa=bbb&key&ccc=ddd\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);
		assert(("key" in parameter) !is null);
		assert(std.string.cmp(parameter["key"][0], "") == 0);
		assert(("ccc" in parameter) !is null);
		assert(std.string.cmp(parameter["ccc"][0], "ddd") == 0);

		parameter = parseQueryString("?aaa=bbb&&=&ccc=ddd\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);
		assert(("" in parameter) !is null);
		assert(std.string.cmp(parameter[""][0], "") == 0);
		assert(("ccc" in parameter) !is null);
		assert(std.string.cmp(parameter["ccc"][0], "ddd") == 0);

		parameter = parseQueryString("?aaa=bbb=ccc&\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb=ccc") == 0);

		parameter = parseQueryString("aaa=bbb&ccc=%E3%81%82%E3%81%84%E3%81%86%E3%81%88%E3%81%8A\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);
		assert(("ccc" in parameter) !is null);
		assert(std.string.cmp(parameter["ccc"][0], "あいうえお") == 0);

		parameter = parseQueryString("aaa=%Ez%8o%8%Exx\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "%Ez%8o%8%Exx") == 0);
	}

	/+
	 + multipart/form-dataを解析
	 +/
	static
	ParsedPair parseMultipartFormData(string boundary, MultipartFormDataCursor cursor)
	{
		ParsedPair pair;

		cursor.skipNextTo("--" ~ boundary);

		boundary = "\r\n--" ~ boundary;
		for (;;)
		{
			if (cursor.startsWith("--"))
				break; // 正常な終端
			else if (cursor.startsWith("\r\n"))
				cursor.skipBytes(2);
			else
			{
				// 異常な終端
			}

			string name;
			string fileName;

			for (;;)
			{
				if (cursor.startsWith("\r\n"))
				{
					// ヘッダ終了
					cursor.skipBytes(2);
					break;
				}
				else if (cursor.startsWith("Content-Type:"))
				{
					// とりあえずイラネ
					cursor.skipNextTo("\r\n");
				}
				else if (cursor.startsWith("Content-Disposition:"))
				{
					cursor.skipBytes(20);

					for (;;)
					{
						cursor.skipNextToSpace();

						if (cursor.startsWith("\r\n"))
						{
							cursor.skipBytes(2);
							break;
						}
						else if (cursor.startsWith("form-data;"))
							cursor.skipBytes(10);
						else if (cursor.startsWith("name"))
						{
							cursor.skipBytes(4);

							cursor.skipNextToSpace();

							if (cursor.startsWith("="))
								cursor.skipBytes(1);
							else
								throw new RequestParseError("Illegal charactor found after 'name'.");

							cursor.skipNextToSpace();

							if (cursor.startsWith("\""))
								cursor.skipBytes(1);
							else
								throw new RequestParseError("Illegal charactor found after 'name='.");

							name = cursor.copyBeforeTo("\"");

							if (cursor.startsWith(";"))
								cursor.skipBytes(1);
						}
						else if (cursor.startsWith("filename"))
						{
							cursor.skipBytes(8);

							cursor.skipNextToSpace();

							if (cursor.startsWith("="))
								cursor.skipBytes(1);
							else
								throw new RequestParseError("Illegal charactor found after 'filename'.");

							cursor.skipNextToSpace();

							if (cursor.startsWith("\""))
								cursor.skipBytes(1);
							else
								throw new RequestParseError("Illegal charactor found after 'filename='.");

							fileName = cursor.copyBeforeTo("\"");

							if (cursor.startsWith(";"))
								cursor.skipBytes(1);
						}
						else
						{
							cursor.skipNextTo("\r\n");
						}
					}
				}
				else
				{
					// 知らないヘッダなので無視する
					cursor.skipNextTo("\r\n");
				}
			}

			if (fileName is null)
			{
				if ((name in pair.parameters) is null)
					pair.parameters[name] = [cursor.copyBeforeTo(boundary)];
				else
					pair.parameters[name] ~= cursor.copyBeforeTo(boundary);
			}
			else
				pair.uploadedFiles[name] = cursor.copyBeforeTo(boundary, fileName);
		}

		return pair;
	}

	/+
	 + Cookieをキーと値の連想配列にパースする
	 +
	 + Params:
	 +  cookieString = パースする文字列 (Cの文字列)
	 +
	 + Returns:
	 +  パースされたデータの連想配列
	 +/
	static
	const(char)[][][const(char)[]] parseCookieString(char* cookieString)
	{
		// 解析結果のデータを格納する
		const(char)[][][const(char)[]] parameter;

		// 入力がなければ空を返す
		if (cookieString is null)
		{
			parameter[null] = null;
			return parameter;
		}

		char[4] multiByteCharactor;
		uint multiByteCharactorLength;

		const(char)[] key;

		uint begin = 0;
		uint read = 0;
		uint write = 0;
		uint status = 0; // 0:key検索中 1:key読み取り中 2:'='検索中 3:value検索中 4:value読み取り中 5:';' or ','検索中

		for (;;)
		{
			char c = cookieString[read++];

			if (c == 0)
				break;

			// key検索中 or value検索中
			if (status == 0 || status == 3)
			{
				if (c != ' ' && c != '\t')
				{
					if (charactorIsControl(c))
						throw new RequestParseError("Cookie string is illegal format.");
					else
					{
						status = status == 0 ? 1 : 4;
						--read;
					}
				}
			}
			// '='検索中
			else if (status == 2)
			{
				if (c == '=')
					status = 3;
				else if (c != ' ' && c != '\t')
					throw new RequestParseError("Cookie string is illegal format.");
			}
			// ';'または','検索中の処理
			else if (status == 5)
			{
				if (c == ';' || c == ',')
					status = 0;
				else if (c != ' ' && c != '\t')
					throw new RequestParseError("Cookie string is illegal format.");
			}
			else
			{
				if (status == 1 && (c == ' ' || c == '\t' || c == '='))
				{
					if (begin == write)
						key = "";
					else
						key = cast(const(char)[])cookieString[begin .. write];
					begin = write;

					if (c == '=')
						status = 3;
					else
						status = 2;
				}
				else if (status == 4 && (c == ' ' || c == '\t' || c == ';' || c == ','))
				{
					if (begin == write)
					{
						if ((key in parameter) is null)
							parameter[key] = [""];
						else
							parameter[key] ~= "";
					}
					else
					{
						if ((key in parameter) is null)
							parameter[key] = [cast(const(char)[])cookieString[begin .. write]];
						else
							parameter[key] ~= cast(const(char)[])cookieString[begin .. write];
						begin = write;
					}

					if (c == ';' || c == ',')
						status = 0;
					else
						status = 5;
				}
				else
				{
					if (charactorIsControl(c))
						throw new RequestParseError("Illegal charactor found in cookie string.");
					else if (c == '%')
					{
						// %が来て、後ろの2文字で16進数が構成できるならばデコードする
						// 後ろ2文字で16進数が構成できない場合は何もしない

						// 0-9か判定
						ubyte upper = cast(ubyte)(cookieString[read] - '0');
						if (upper < 10)
							upper *= 16;
						else
						{
							// A-Fか判定
							upper -= cast(ubyte)'A' - cast(ubyte)'0';
							if (upper < 6)
							{
								upper += 10;
								upper <<= 4;
							}
							else
							{
								// a-fか判定
								upper -= cast(ubyte)'a' - cast(ubyte)'A';
								if (upper < 6)
								{
									upper += 10;
									upper <<= 4;
								}
								else
								{
									// 16進数でないのでエンコードされていないと判定
									cookieString[write++] = c;
									multiByteCharactorLength = 0;
									continue;
								}
							}
						}

						// 0-9か判定
						ubyte lower = cast(ubyte)(cookieString[read+1] - '0');
						if (lower < 10)
							upper += lower;
						else
						{
							// A-Fか判定
							lower -= cast(ubyte)'A' - cast(ubyte)'0';
							if (lower < 6)
								upper += lower + 10;
							else
							{
								// a-fか判定
								lower -= cast(ubyte)'a' - cast(ubyte)'A';
								if (lower < 6)
									upper += lower + 10;
								else
								{
									// 16進数でないのでエンコードされていないと判定
									cookieString[write++] = c;
									multiByteCharactorLength = 0;
									continue;
								}
							}
						}

						// マルチバイト文字か判定
						if (upper >= 0x80)
						{
							if (multiByteCharactorLength == 0)
							{
								// UTF-8マルチバイト文字1文字目として正当か判定
								if ((upper & 0b11100000) == 0b11000000 || (upper & 0b11110000) == 0b11100000 || (upper & 0b11111000) == 0b11110000)
								{
									multiByteCharactor[multiByteCharactorLength++] = cast(char)upper;
								}
							}
							else
							{
								// UTF-8マルチバイト文字2文字目以降として正当か判定
								if ((upper & 0b11000000) == 0b10000000)
								{
									multiByteCharactor[multiByteCharactorLength++] = cast(char)upper;
									if (multiByteCharactorLength == 2 && (multiByteCharactor[0] & 0b11100000) == 0b11000000)
									{
										cookieString[write++] = multiByteCharactor[0];
										cookieString[write++] = multiByteCharactor[1];

										multiByteCharactorLength = 0;
									}
									else if (multiByteCharactorLength == 3 && (multiByteCharactor[0] & 0b11110000) == 0b11100000)
									{
										cookieString[write++] = multiByteCharactor[0];
										cookieString[write++] = multiByteCharactor[1];
										cookieString[write++] = multiByteCharactor[2];

										multiByteCharactorLength = 0;
									}
									else if (multiByteCharactorLength == 4)
									{
										cookieString[write++] = multiByteCharactor[0];
										cookieString[write++] = multiByteCharactor[1];
										cookieString[write++] = multiByteCharactor[2];
										cookieString[write++] = multiByteCharactor[3];

										multiByteCharactorLength = 0;
									}
								}
								else
								{
									multiByteCharactorLength = 0;
								}
							}
						}
						else
						{
							// 途中で切れたら不正なので無かったことにする
							multiByteCharactorLength = 0;

							// 不正な文字でなかったら書き込む
							if (cast(ubyte)((upper ^ 0x40) - 0x3f) > 0x20 || upper == 0x09 || upper == 0x0a || upper == 0x0d)
								cookieString[write++] = cast(char)upper;
						}

						read += 2;
					}
					// '+'は半角スペースに変換する
					else if (c == '+')
					{
						cookieString[write++] = ' ';
						multiByteCharactorLength = 0;
					}
					else
					{
						cookieString[write++] = c;
						multiByteCharactorLength = 0;
					}
				}
			}
		}

		// 最後に読み込み中であったものを処理する
		if (status == 1)
		{
			if (begin < write)
			{
				const(char)[] k = cast(const(char)[])cookieString[begin .. write];
				if ((k in parameter) is null)
					parameter[k] = [""];
				else
					parameter[k] ~= "";
			}
		}
		else if (status == 2 || status == 3)
		{
			if ((key in parameter) is null)
				parameter[key] = [""];
			else
				parameter[key] ~= "";
		}
		else if (status == 4)
		{
			if (begin == write)
			{
				if ((key in parameter) is null)
					parameter[key] = [""];
				else
					parameter[key] ~= "";
			}
			else
			{
				if ((key in parameter) is null)
					parameter[key] = [cast(const(char)[])cookieString[begin .. write]];
				else
					parameter[key] ~= cast(const(char)[])cookieString[begin .. write];
			}
		}

		return parameter;
	}
	unittest
	{
		// パラメータのパースをテスト
		const(char)[][][const(char)[]] parameter;

		parameter = parseCookieString("aaa=bbb\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);

		parameter = parseCookieString(" aaa = bbb ; \0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);

		parameter = parseCookieString(" aaa = bbb , \0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);

		parameter = parseCookieString("aaa=bbb; ccc=ddd\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);
		assert(("ccc" in parameter) !is null);
		assert(std.string.cmp(parameter["ccc"][0], "ddd") == 0);

		parameter = parseCookieString("aaa=bbb, ccc=ddd,\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);
		assert(("ccc" in parameter) !is null);
		assert(std.string.cmp(parameter["ccc"][0], "ddd") == 0);

		parameter = parseCookieString("ccc=%E3%81%82%E3%81%84%E3%81%86%E3%81%88%E3%81%8A , aaa=bbb ;\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "bbb") == 0);
		assert(("ccc" in parameter) !is null);
		assert(std.string.cmp(parameter["ccc"][0], "あいうえお") == 0);

		parameter = parseCookieString("aaa=%Ez%8o%8%Exx\0".dup.ptr);
		assert(("aaa" in parameter) !is null);
		assert(std.string.cmp(parameter["aaa"][0], "%Ez%8o%8%Exx") == 0);
	}

	/+
	 + 文字が制御文字か判定する
	 +
	 + 0x80以降も制御文字と見なす
	 +
	 + Params:
	 +  c = 判定する文字
	 +
	 + Returns:
	 +  制御文字だった場合にtrueを返す
	 +/
	static pure
	bool charactorIsControl(char c)
	{
		return cast(ubyte)(c - 0x20) >= 0x5f;
	}

	/+
	 + parsemultipartformdataの戻り値用 Tupleがなんかコンパイルできねんで・・・
	 +/
	static
	struct ParsedPair
	{
		const(char)[][][const(char)[]] parameters;
		UploadedFile[string] uploadedFiles;
	}

	/+
	 + multipart/form-dataを解析するときに使用するクラス
	 +
	 + 目印になる記号が複数の文字だったり、
	 + 読み込みもとがストリーム(たぶん実体はファイル)だったりするので、
	 + 割と複雑な処理が必要になってしまう。
	 + そのあたりの面倒な処理を色々いれたクラス
	 +/
	static
	class MultipartFormDataCursor
	{
		/+
		 + コンストラクタ
		 +
		 + Params:
		 +  buffer = 解析に使用するバッファ (長さは2の累乗 1はだめ)
		 +  onMemoryThreshold = アップロードされたファイルをメモリに乗せる閾値
		 +  directory = アップロードされたファイルをディスクに書き出す際のディレクトリ
		 +/
		this(ubyte[] buffer, uint onMemoryThreshold, string directory)
		in
		{
			bool flag = false;
			for (uint i = 1; i < 32; ++i)
			{
				if ((1 << i) == buffer.length)
				{
					flag = true;
					break;
				}
			}
			assert(flag);
		}
		body
		{
			buffer_ = buffer;
			onMemoryThreshold_ = onMemoryThreshold;
			directory_ = directory;
		}

		/+
		 + 読み出すストリームを設定
		 +
		 + Params:
		 +  stream = データを読む対象のストリーム
		 +/
		void initialize(FCGX_Stream* stream)
		{
			streamReadComplete_ = false;
			stream_ = stream;
			cursor_ = 0;
			readSize_ = 0;
		}

		/+
		 + 指定した文字列を検索する
		 +
		 + カーソル位置から開始して、指定文字列を検索する
		 + 検索終了後のカーソル位置は、見つかった文字列の次の文字
		 +
		 + Params:
		 +  source = 検索文字列
		 +
		 + Throws:
		 +  終端まで検索しても見つからない場合
		 +/
		void skipNextTo(string source)
		in
		{
			assert(source.length <= buffer_.length / 2);
		}
		body
		{
			for (;;)
			{
				// バッファにデータを補充
				autoFillNextHalfBuffer();

				// 未処理データサイズが検索文字列サイズより小さかったらエラー
				if (readSize_ - cursor_ < source.length)
					throw new RequestParseError("Can not found string '" ~ source ~ "'.");

				// 検索しても平気なサイズを計算
				uint length = readSize_ - cursor_ - (cast(uint)source.length - 1);

				// 検索処理
				const(ubyte)[] buffer = buffer_;
				uint position = cursor_;
				uint mask = cast(uint)buffer_.length - 1;
				ubyte top = source[0];
				for (uint i = 0; i < length; ++i)
				{
					if (buffer[(position + i) & mask] == top)
					{
						uint j;
						for (j = 1; j < source.length; ++j)
						{
							if (buffer[(position + i + j) & mask] != source[j])
								break;
						}

						if (j == source.length)
						{
							cursor_ += i + j;
							return;
						}
					}
				}

				cursor_ += length;
			}
		}

		/+
		 + カーソルを空白の次に移動する
		 +
		 + 現在のカーソル位置が空白だったら、次に空白以外が来るところまでスキップする
		 + 空白と判定するのは半角スペースと水平タブ
		 +
		 + Throws:
		 +  終端まで検索しても見つからない場合
		 +/
		void skipNextToSpace()
		{
			for (;;)
			{
				// バッファにデータを補充
				autoFillNextHalfBuffer();

				// 検索しても平気なサイズを計算
				uint length = readSize_ - cursor_;
				if (length == 0) 
					throw new RequestParseError("Can not found space charactor.");

				// 検索処理
				const(ubyte)[] buffer = buffer_;
				uint position = cursor_;
				uint mask = cast(uint)buffer_.length - 1;
				for (uint i = 0; i < length; ++i)
				{
					ubyte charactor = buffer[(position + i) & mask];
					if (charactor != 0x20 && charactor != 0x09)
					{
						cursor_ += i;
						return;
					}
				}

				cursor_ += length;
			}
		}


		/+
		 + カーソルを指定byte分進める
		 +
		 + Params:
		 +  length = 進めるbyte数
		 +/
		void skipBytes(uint length)
		{
			for (;;)
			{
				// バッファにデータを補充
				autoFillNextHalfBuffer();

				if (readSize_ - cursor_ >= length)
				{
					// 未処理バッファがスキップ数以上なら普通にスキップ
					cursor_ += length;
					break;
				}
				else
				{
					// とりあえずあるだけスキップ
					cursor_ = readSize_;
					length -= readSize_ - cursor_;

					// もう読めなかったらだめ
					if (streamReadComplete_)
						throw new RequestParseError("Can not skip " ~ std.conv.to!string(length) ~ " byte.");
				}
			}
		}

		/+
		 + 現在のカーソル位置の文字列が、指定の文字列と一致するか確認
		 +
		 + Params:
		 +  source = 比較対象の文字列
		 +
		 + Returns:
		 +  一致した場合にtrueを返す
		 +
		 + Throws:
		 +  RequestParseError = 未処理のデータが検索文字列より短い場合
		 +/
		bool startsWith(string source)
		{
			// バッファにデータを補充
			autoFillNextHalfBuffer();

			// 未処理データサイズが検索文字列サイズより小さかったらエラー
			if (readSize_ - cursor_ < source.length)
				throw new RequestParseError("String '" ~ source ~ "' is too long.");

			// 一致を確認
			const(ubyte)[] buffer = buffer_;
			uint position = cursor_;
			uint mask = cast(uint)buffer_.length - 1;
			for (uint i = 0; i < source.length; ++i)
			{
				if (buffer[(position + i) & mask] != source[i])
					return false;
			}

			return true;
		}

		/+
		 + カーソル位置から、指定文字列直前までのデータをコピーして返す
		 +
		 + 副作用として、カーソル位置が指定文字列の次の文字まで移動する
		 +
		 + Params:
		 +  source = 検索対象の文字列
		 +
		 + Returns:
		 +  コピーしたデータ
		 +
		 + Throws:
		 +  RequestParseError = ストリーム終端までに指定文字列が現れない場合
		 +/
		string copyBeforeTo(string source)
		{
			ubyte[] result;
			for (;;)
			{
				// バッファにデータを補充
				autoFillNextHalfBuffer();

				// 未処理データサイズが検索文字列サイズより小さかったらエラー
				if (readSize_ - cursor_ < source.length)
					throw new RequestParseError("Can not found string '" ~ source ~ "'.");

				// 検索しても平気なサイズを計算
				uint length = readSize_ - cursor_ - (cast(uint)source.length - 1);

				// 検索処理
				const(ubyte)[] buffer = buffer_;
				uint position = cursor_;
				uint mask = cast(uint)buffer_.length - 1;
				ubyte top = source[0];
				for (uint i = 0; i < length; ++i)
				{
					if (buffer[(position + i) & mask] == top)
					{
						uint j;
						for (j = 1; j < source.length; ++j)
						{
							if (buffer[(position + i + j) & mask] != source[j])
								break;
						}

						if (j == source.length)
						{
							// 見つかった文字列の直前までのデータをファイルに追加
							if (i > 0)
							{
								uint start = position & mask;
								uint end = (position + i) & mask;
								if (start < end)
									result ~= buffer[start .. end];
								else
								{
									result ~= buffer[start .. $];
									if (end > 0)
										result ~= buffer[0 .. end];
								}
							}

							cursor_ += i + j;

							if (result is null)
								return "";
							else
								return cast(string)result;
						}
					}
				}

				// 検索して見つからなかった分を追加
				uint start = position & mask;
				uint end = (position + length) & mask;
				if (start < end)
					result ~= buffer[start .. end];
				else
				{
					result ~= buffer[start .. $];
					if (end > 0)
						result ~= buffer[0 .. end];
				}

				cursor_ += length;
			}
		}
	
		/+
		 + カーソル位置から、指定文字列直前までのデータをコピーして返す
		 +
		 + 副作用として、カーソル位置が指定文字列の次の文字まで移動する
		 +
		 + Params:
		 +  source = 検索対象の文字列
		 +
		 + Returns:
		 +  コピーしたデータ
		 +
		 + Throws:
		 +  RequestParseError = ストリーム終端までに指定文字列が現れない場合
		 +/
		UploadedFile copyBeforeTo(string source, string fileName)
		{
			auto file = new UploadedFile(fileName, onMemoryThreshold_, directory_);
			scope(exit) file.closeFileIfOpen();

			for (;;)
			{
				// バッファにデータを補充
				autoFillNextHalfBuffer();

				// 未処理データサイズが検索文字列サイズより小さかったらエラー
				if (readSize_ - cursor_ < source.length)
					throw new RequestParseError("Can not found string '" ~ source ~ "'.");

				// 検索しても平気なサイズを計算
				uint length = readSize_ - cursor_ - (cast(uint)source.length - 1);

				// 検索処理
				const(ubyte)[] buffer = buffer_;
				uint position = cursor_;
				uint mask = cast(uint)buffer_.length - 1;
				ubyte top = source[0];
				for (uint i = 0; i < length; ++i)
				{
					if (buffer[(position + i) & mask] == top)
					{
						uint j;
						for (j = 1; j < source.length; ++j)
						{
							if (buffer[(position + i + j) & mask] != source[j])
								break;
						}

						if (j == source.length)
						{
							// 見つかった文字列の直前までのデータをファイルに追加
							if (i > 0)
							{
								uint start = position & mask;
								uint end = (position + i) & mask;
								if (start < end)
									file.append(buffer[start .. end]);
								else
								{
									file.append(buffer[start .. $]);
									if (end > 0)
										file.append(buffer[0 .. end]);
								}
							}

							cursor_ += i + j;

							return file;
						}
					}
				}

				// 検索して見つからなかった分をファイルに追加
				uint start = position & mask;
				uint end = (position + length) & mask;
				if (start < end)
					file.append(buffer[start .. end]);
				else
				{
					file.append(buffer[start .. $]);
					if (end > 0)
						file.append(buffer[0 .. end]);
				}

				cursor_ += length;
			}
		}

	private:

		// ストリーム読み込み完了フラグ
		bool streamReadComplete_;

		// 読み込み対象のストリーム
		FCGX_Stream* stream_;

		// 現在の処理位置
		uint cursor_;

		// バッファに読み込んだ総量
		uint readSize_;

		// バッファ
		ubyte[] buffer_;

		// アップロードされたファイルをメモリに乗せる時の最大値
		uint onMemoryThreshold_;

		// アップロードされたファイルを保存するディレクトリ
		string directory_;

		/+
		 + バッファにデータを読み出す
		 +
		 + バッファ上に未処理のデータが、半分より少なければ
		 + ストリームからバッファの半分のサイズのデータを読み出す
		 + ストリームから読み取れない場合は当然補充できないので
		 + 必ず半分以上未処理データが残るという訳ではない
		 +/
		void autoFillNextHalfBuffer()
		{
			// 読み込み完了してたら無理
			if (streamReadComplete_)
				return;

			// 既に十分読み込んでいれば何もしない
			uint length = cast(uint)(buffer_.length / 2);
			if (readSize_ - cursor_ > length)
				return;

			// バッファの書き込み位置と書き込む量を計算
			char* start = cast(char*)&buffer_[(readSize_ + ($ / 2 - 1)) & ($ / 2)];

			// ストリームからバッファに書き込む
			int size = FCGX_GetStr(start, length, stream_); 
			if (size < length)
				streamReadComplete_ = true;

			readSize_ += size;
		}
	}
}


/++
 + クライアントに返すレスポンスクラス
 +/
class Response
{
	/// HTTP status code
	enum StatusCode
	{
		CONTINUE,
		SWITCHING_PROTOCOLS,
		OK,
		CREATED,
		ACCEPTED,
		NON_AUTHORITATIVE_INFORMATION,
		NO_CONTENT,
		RESET_CONTENT,
		PARTIAL_CONTENT,
		MULTIPLE_CHOICES,
		MOVED_PERMANENTLY,
		FOUND,
		SEE_OTHER,
		NOT_MODIFIED,
		USE_PROXY,
		TEMPORARY_REDIRECT,
		BAD_REQUEST,
		UNAUTHORIZED,
		PAYMENT_REQUIRED,
		FORBIDDEN,
		NOT_FOUND,
		METHOD_NOT_ALLOWED,
		NOT_ACCEPTABLE,
		PROXY_AUTHENTICATION_REQUIRED,
		REQUEST_TIME_OUT,
		CONFLICT,
		GONE,
		LENGTH_REQUIRED,
		PRECONDITION_FAILED,
		REQUEST_ENTITY_TOO_LARGE,
		REQUEST_URI_TOO_LARGE,
		UNSUPPORTED_MEDIA_TYPE,
		REQUESTED_RANGE_NOT_SATISFIABLE,
		EXPECTATION_FAILED,
		INTERNAL_SERVER_ERROR,
		NOT_IMPLEMENTED,
		BAD_GATEWAY,
		SERVICE_UNAVAILABLE,
		GATEWAY_TIME_OUT,
		HTTP_VERSION_NOT_SUPPORTED,
	}

	/// ヘッダの"Status:"の値
	StatusCode statusCode;

	/// ヘッダの"Content-Type:"の値
	string contentType;

	/// 任意追加のヘッダ
	string[string] optionalHeaders;

	/// リダイレクト先URL これを指定するとLocationヘッダのみを出力し、他は無視する
	string redirect;

	/// リクエスト処理中に発生した例外
	Throwable thrownObject;

	/++
	 + コンストラクタ
	 +/
	this()
	{
		buffer_ = cast(char*)core.stdc.stdlib.malloc(8 * 1024);
		headerLength_ = 1024;

		reset();
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		core.stdc.stdlib.free(buffer_);
	}

	/++
	 + リクエスト処理前に内部データをリセットする
	 +/
	void reset()
	{
		statusCode = StatusCode.OK;
		contentType = null;
		optionalHeaders = null;
		redirect = null;
		thrownObject = null;
		headerWriteCursor_ = headerLength_ - 1;
		contentWriteCursor_ = headerLength_;
	}

	/++
	 + 設定されたデータをクライアントに返す
	 +/
	void output(FCGX_Stream* stream, Request request)
	{
		// リダイレクト先が指定されていたらリダイレクトする
		if (redirect !is null)
		{
			string redirectUrl = redirect;

			// プロトコルの指定がなければ付ける
			if (std.string.indexOf(redirectUrl, "://") < 0)
			{
				// 相対パスだったら絶対パスにする
				if (!std.string.startsWith(redirectUrl, "/"))
				{
					// 環境変数からSCRIPT_NAMEを取得
					string path = request.getEnvironmentVariable("SCRIPT_NAME");

					// SCRIPT_NAME中のディレクトリパス終端を取得
					ptrdiff_t endOfDirectory;
					if (path !is null)
						endOfDirectory = std.string.lastIndexOf(path, "/");
					else
						endOfDirectory = -1;

					// 相対パスを付けて絶対パスにする(..とか.とかもそのまま付けちゃう)
					if (endOfDirectory < 0)
						redirectUrl = "/" ~ redirectUrl;
					else
						redirectUrl = path[0 .. endOfDirectory + 1] ~ redirectUrl;
				}

				// プロトコルを判定
				string protocol = request.getEnvironmentVariable("HTTPS");
				if (protocol !is null && std.string.cmp(protocol, "on") == 0)
					protocol = "https://";
				else
					protocol = "http://";

				// サーバ名を取得
				string serverName;
				foreach (variable; ["SERVER_NAME", "HTTP_HOST", "REMOTE_HOST"])
				{
					serverName = request.getEnvironmentVariable(variable);
					if (serverName !is null && serverName.length > 0)
						break;
				}

				if (serverName !is null && serverName.length > 0)
					redirectUrl = protocol ~ serverName ~ redirectUrl;
			}

			string location = "Location: " ~ redirectUrl ~ "\r\n";

			// 任意追加のヘッダを出力
			if (optionalHeaders !is null)
			{
				// 特にエスケープはしない
				foreach (key, value; optionalHeaders)
					location ~= key ~ ": " ~ value ~ "\r\n";
			}

			location ~= "\r\n";

			FCGX_PutStr(location.ptr, cast(int)location.length, stream);
		}
		else
		{
			// ヘッダと本文の区切り
			appendHeader("\r\n");

			// 任意追加のヘッダを出力
			if (optionalHeaders !is null)
			{
				// 特にエスケープはしない
				foreach (key, value; optionalHeaders)
				{
					appendHeader("\r\n");
					appendHeader(value);
					appendHeader(": ");
					appendHeader(key);
				}
			}

			// Content-Lengthヘッダを出力
			appendHeader("\r\n");
			appendHeader(std.conv.to!string(contentWriteCursor_ - headerLength_));
			appendHeader("Content-Length: ");

			// Content-Typeヘッダを出力
			if (contentType is null)
			{
				// 無いなら仕方ないので適当に出しておく
				appendHeader("Content-Type: text/plain\r\n");
			}
			else
			{
				appendHeader("\r\n");
				appendHeader(contentType);
				appendHeader("Content-Type: ");
			}

			// HTTPのステータスコードを出力 (200 OKの場合は省略)
			if (statusCode != StatusCode.OK)
				appendHeader(STATUS_CODE_CGI_HEADERS[statusCode]);

			// データを出力
			FCGX_PutStr(&buffer_[headerWriteCursor_+1], contentWriteCursor_ - (headerWriteCursor_ + 1), stream);
		}
	}

	/++
	 + 出力バッファにコンテンツを追加
	 +
	 + Param:
	 +  data = 追加するデータ
	 +/
	void append(const(char)[] data)
	{
		// バッファが足りなければ拡張する
		uint length = cast(uint)data.length;
		uint newBufferSize = bufferSize_;
		while (contentWriteCursor_ + length > newBufferSize)
			newBufferSize = newBufferSize * 2 + 1;

		if (newBufferSize != bufferSize_)
		{
			buffer_ = cast(char*)core.stdc.stdlib.realloc(buffer_, newBufferSize);
			bufferSize_ = newBufferSize;
		}

		// 書き込み
		const(char)* read = data.ptr;
		char* write = &buffer_[contentWriteCursor_];
		for (uint i = 0; i < length; ++i)
			write[i] = read[i];
		contentWriteCursor_ += data.length;
	}

private:

	// 各Status codeに対応するヘッダ文字列
	static immutable
	string[] STATUS_CODE_CGI_HEADERS =
	[
		"Status: 100 Continue\r\n",
		"Status: 101 Switching Protocols\r\n",
		"Status: 200 OK\r\n",
		"Status: 201 Created\r\n",
		"Status: 202 Accepted\r\n",
		"Status: 203 Non-Authoritative Information\r\n",
		"Status: 204 No Content\r\n",
		"Status: 205 Reset Content\r\n",
		"Status: 206 Partial Content\r\n",
		"Status: 300 Multiple Choices\r\n",
		"Status: 301 Moved Permanently\r\n",
		"Status: 302 Found\r\n",
		"Status: 303 See Other\r\n",
		"Status: 304 Not Modified\r\n",
		"Status: 305 Use Proxy\r\n",
		"Status: 307 Temporary Redirect\r\n",
		"Status: 400 Bad Request\r\n",
		"Status: 401 Unauthorized\r\n",
		"Status: 402 Payment Required\r\n",
		"Status: 403 Forbidden\r\n",
		"Status: 404 Not Found\r\n",
		"Status: 405 Method Not Allowed\r\n",
		"Status: 406 Not Acceptable\r\n",
		"Status: 407 Proxy Authentication Required\r\n",
		"Status: 408 Request Time-out\r\n",
		"Status: 409 Conflict\r\n",
		"Status: 410 Gone\r\n",
		"Status: 411 Length Required\r\n",
		"Status: 412 Precondition Failed\r\n",
		"Status: 413 Request Entity Too Large\r\n",
		"Status: 414 Request-URI Too Large\r\n",
		"Status: 415 Unsupported Media Type\r\n",
		"Status: 416 Requested range not satisfiable\r\n",
		"Status: 417 Expectation Failed\r\n",
		"Status: 500 Internal Server Error\r\n",
		"Status: 501 Not Implemented\r\n",
		"Status: 502 Bad Gateway\r\n",
		"Status: 503 Service Unavailable\r\n",
		"Status: 504 Gateway Time-out\r\n",
		"Status: 505 HTTP Version not supported\r\n",
	];

	// 出力データバッファ
	char* buffer_;

	// 出力データバッファサイズ
	uint bufferSize_;

	// 出力データバッファ中のヘッダー部分の長さ
	uint headerLength_;

	// 出力バッファ中のヘッダー書き込み位置
	uint headerWriteCursor_;

	// 出力バッファ中のコンテンツ書き込み位置
	uint contentWriteCursor_;

	/+
	 + 出力バッファにヘッダ内容を追加
	 +
	 + Params:
	 +  data = 追加する内容
	 +/
	void appendHeader(const(char)[] data)
	{
		if (data is null || data.length == 0)
			return;

		// 残りのバッファが無いときの処理
		if (headerWriteCursor_ + 1 < data.length)
		{
			// 全体の残りバッファを確認して、足りなければ伸ばす
			uint newBufferSize = bufferSize_;
			while ((headerWriteCursor_ + 1) + (newBufferSize - contentWriteCursor_) < data.length)
				newBufferSize = newBufferSize * 2 + 1;

			if (newBufferSize != bufferSize_)
			{
				buffer_ = cast(char*)core.stdc.stdlib.realloc(buffer_, newBufferSize);
				bufferSize_ = newBufferSize;
			}

			// ヘッダが入るように後ろにずらす
			uint shift = cast(uint)data.length - (headerWriteCursor_ + 1);
			uint end = headerWriteCursor_;
			char* buffer = buffer_;
			for (uint i = contentWriteCursor_ - 1; i > end; --i)
				buffer[i + shift] = buffer[i];

			// カーソル位置もずらす
			headerWriteCursor_ += shift;
			contentWriteCursor_ += shift;

			// 次から入るように元のサイズを拡張する
			headerLength_ += shift;
		}

		// データを追加
		const(char)* read = data.ptr;
		char* write = &buffer_[(headerWriteCursor_ + 1) - data.length];
		for (uint i = 0; i < data.length; ++i)
			write[i] = read[i];
		headerWriteCursor_ -= data.length;
	}
}

/++
 + フィルタ
 +
 + フィルタはこのクラスを継承し、executeメソッドを実装した上で、
 + RaptanのaddFilterメソッドにインスタンスを投入する
 +/
class Filter
{
	/++
	 + フィルタの処理
	 +
	 + 後続のフィルタを呼び出す場合はexecuteAfterFiltersメソッドを使用する
	 + フィルタリストの最後に設置されたフィルタでActionの呼び出しが行われるので、
	 + Action前後の処理の定義は下記のようになる。
	 +
	 + executeAfterFiltersメソッド以前の処理がAction実行前の処理
	 + executeAfterFiltersメソッド以後の処理がAction実行後の処理
	 +
	 + executeAfterFiltersメソッドを呼び出さないと、以降のフィルタ処理や
	 + Action呼び出しが行われない
	 +
	 + Params:
	 +  request = リクエスト
	 +  response = レスポンス
	 +/
	abstract
	void execute(Request request, Response response);

protected:

	/++
	 + 後続のフィルタを実行
	 +
	 + Params:
	 +  request = リクエスト
	 +  response = レスポンス
	 +/
	final
	void executeAfterFilters(Request request, Response response)
	{
		nextFilter_.execute(request, response);
	}

private:

	// 次のフィルタ
	Filter nextFilter_;
}

/++
 + FastCGI用フレームワーク
 +/
class Raptan : Thread
{
	/// リクエストを実装するDelegateの定義
	alias string delegate(Request, Response) ActionDelegate;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  bufferSize = POSTされたデータの解析用バッファ
	 +  uploadFileOnMemoryThreshold = アップロードされたファイルをメモリに乗せる時の最大値 0を指定すると常にファイルに保存する
	 +  uploadDirectory = アップロードされたファイルを保存するディレクトリ
	 +/
	this(uint bufferSize, uint uploadFileOnMemoryThreshold, string uploadDirectory)
	{
		synchronized
		{
			if (!initialize)
			{
				FCGX_Init();
				initialize = true;
			}
		}

		request_ = new Request(bufferSize, uploadFileOnMemoryThreshold, uploadDirectory);
		response_ = new Response;
		filters_ ~= new DispatchActionFilter;

		super(&run);
	}

	/++
	 + URIに対するアクションを設定
	 +
	 + Params:
	 +  uri = URI
	 +  action = アクション
	 +/
	void setAction(string uri, ActionDelegate action)
	{
		actions_[uri] = action;
	}

	/++
	 + アクション実行前に実行されるフィルタを設定
	 +
	 + フィルタは登録された順と逆順に実行される
	 + Actionが底のスタックにaddで積み上げるイメージ
	 +
	 + Params:
	 +  filter = フィルタ
	 +/
	void addFilter(Filter filter)
	{
		filter.nextFilter_ = filters_[$ - 1];
		filters_ ~= filter;
	}

	/++
	 + リクエストされたURIに対するDelegateが設定されてない場合の処理を設定
	 +
	 + ここで設定したdelegateが文字列を返しても無視します
	 +
	 + Params:
	 +  action = 処理
	 +/
	void notFoundAction(ActionDelegate action)
	{
		notFoundAction_ = action;
	}

	/++
	 + 例外が発生した場合の処理を設定
	 +
	 + フィルタ処理中の例外も受け取るので
	 + ここで設定するActionに限り、
	 + フィルタ処理が実行されることを前提としてはならない
	 + また、ここで設定したdelegateが文字列を返しても無視します
	 +
	 + Params:
	 +  action = 処理
	 +/
	void internalServerErrorAction(ActionDelegate action)
	{
		internalServerErrorAction_ = action;
	}

	/++
	 + 実行関数
	 +/
	void run()
	{
		while (request_.accept())
		{
			// 新たな接続のためにレスポンスをリセットする
			response_.reset();

			try
			{
				filters_[$ - 1].execute(request_, response_);
			}
			catch (Throwable t)
			{
				// 例外が発生した場合の処理
				response_.thrownObject = t;
				if (internalServerErrorAction_ is null)
				{
					response_.statusCode = Response.StatusCode.INTERNAL_SERVER_ERROR;
					response_.contentType = null;
					response_.redirect = null;
				}
				else
				{
					internalServerErrorAction_(request_, response_);
				}
			}

			// レスポンスを出力
			response_.output(request_.fcgxRequest_.output, request_);
		}
	}

private:

	// FCGX初期化フラグ
	static
	bool initialize = false;

	// リクエスト
	Request request_;

	// レスポンス
	Response response_;

	// フィルタのリスト
	Filter[] filters_;

	// アクションのマップ
	ActionDelegate[string] actions_;

	// リクエストされたURLに対するDelegateが設定されていない場合の処理
	ActionDelegate notFoundAction_;

	// アクションを処理する際にエラーが発生した場合の処理を設定
	ActionDelegate internalServerErrorAction_;

	/+
	 + リクエストURIを判定して各アクションに振り分けるフィルタ
	 +/
	class DispatchActionFilter : Filter
	{
		override
		void execute(Request request, Response response)
		{
			string uri = request.uri;
			while (uri !is null)
			{
				ActionDelegate* p = uri in actions_;
				if (p is null)
				{
					// リクエストされたURIに対するActionが存在しない場合の処理
					if (notFoundAction_ is null)
					{
						response.statusCode = Response.StatusCode.NOT_FOUND;
						response.contentType = null;
						response.redirect = null;
					}
					else
					{
						notFoundAction_(request, response);
					}
					return;
				}
				else
				{
					// Actionを実行
					uri = (*p)(request, response);
				}
			}
		}
	}
}

