/++
 + テンプレートエンジン スイーツ(笑) Version: 0.4.1
 +
 + フォーム制御モジュール
 +
 + Copyright(C) 2009-2011 Takuya Kurosawa
 +/
module jp.ku6.sweets.form;

import std.conv;
import std.regex;
import std.array;
import std.string;

import jp.ku6.sweets.core;
import jp.ku6.sweets.extension;
import jp.ku6.sweets.utility;
import jp.ku6.raptan.core;

/++
 + メッセージを指定されたプレースホルダで分割する
 +
 + 分割したメッセージはプレースホルダも含む
 + messageBuild関数ででの使用を想定している
 +
 + Params:
 +  message = 対象のメッセージ
 +  placeHolder = プレースホルダの配列
 +
 + Return:
 +  分割されたメッセージ
 +/
string[] messageSplit(string message, string[] placeHolders)
{
	string[] result;
	if (message is null || message.length == 0)
		return result;

	uint start = 0;
	for (uint i = 0; i < message.length; ++i)
	{
		foreach (placeHolder; placeHolders)
		{
			if (message.length - i < placeHolder.length)
				continue;

			// 一致する文字数をチェック
			uint match = 0;
			for (; match < placeHolder.length && message[i+match] == placeHolder[match]; ++match) {}

			if (match == placeHolder.length)
			{
				if (start < i)
					result ~= message[start .. i];

				// 結果を保存
				result ~= placeHolder;

				// プレースホルダの分スキップする
				i += placeHolder.length;

				// 出発点を再設定
				start = i;

				// ++iされるので引いておく
				--i;

				break;
			}
		}
	}

	// 残りをはき出す
	if (start < message.length - 1)
		result ~= message[start .. $];

	return result;
}


/++
 + 分割されたメッセージを再構築する
 +
 + 分割されたメッセージと、プレースホルダに対する置き換え文字列を指定し
 + 一つのメッセージに再構築する
 +
 + Params:
 +  parts = 分割されたメッセージ
 +  replace = 置き換え文字列 プレースホルダがキー、置き換え文字列が値の連想配列
 +
 + Returns;
 +  再構築されたメッセージ
 +/
string messageBuild(string[] parts, string[string] replace)
{
	// メッセージ無い場合
	if (parts is null || parts.length == 0)
		return "";

	// 置換しなくて良い場合
	if (replace is null || replace.length == 0)
	{
		// 1個だけの時は連結しなくてよい
		if (parts.length == 1)
			return parts[0];
		// 2個以上の時は連結する
		else
		{
			// まずは生成される文字列の長さを取得
			size_t length;
			foreach (part; parts)
				length += part.length;

			// メモリを確保
			char[] result = new char[length];
			uint write = 0;

			// 確保したメモリにコピー
			foreach (part; parts)
			{
				const(char)* read = part.ptr;
				length = part.length;
				for (size_t i; i < length; ++i)
					result[write++] = read[i];
			}

			return cast(string)result;
		}
	}
	// 置換する場合
	else
	{
		// 1個だけの時は連結しなくてよい
		if (parts.length == 1)
		{
			string* p = parts[0] in replace;
			if (p is null)
				return parts[0];
			else
				return *p;
		}
		// 2個以上の時は連結する
		else
		{
			// まずは生成される文字列の長さを取得
			size_t length;
			foreach (part; parts)
			{
				string* p = part in replace;
				if (p is null)
					length += part.length;
				else
					length += (*p).length;
			}

			// メモリを確保
			char[] result = new char[length];
			uint write = 0;

			// 確保したメモリにコピー
			foreach (part; parts)
			{
				const(char)* read;

				string* p = part in replace;
				if (p is null)
				{
					read = part.ptr;
					length = part.length;
				}
				else
				{
					read = (*p).ptr;
					length = (*p).length;
				}

				for (size_t i; i < length; ++i)
					result[write++] = read[i];
			}

			return cast(string)result;
		}
	}
}
unittest
{
	string[] placeHolders = ["_LABEL_", "_VALUE_", "_MAXLENGTH_"];

	string[string] replace;
	replace["_LABEL_"] = "ラベル";
	replace["_VALUE_"] = "値";
	replace["_MAXLENGTH_"] = "長さ";

	string[] result;

	result = messageSplit("_LABEL_に_VALUE_文字入力されました。_MAXLENGTH_以内で入力してください。", placeHolders);
	assert(std.string.cmp(result[0], "_LABEL_") == 0);
	assert(std.string.cmp(result[1], "に") == 0);
	assert(std.string.cmp(result[2], "_VALUE_") == 0);
	assert(std.string.cmp(result[3], "文字入力されました。") == 0);
	assert(std.string.cmp(result[4], "_MAXLENGTH_") == 0);
	assert(std.string.cmp(result[5], "以内で入力してください。") == 0);

	assert(std.string.cmp(messageBuild(result, replace), "ラベルに値文字入力されました。長さ以内で入力してください。") == 0);

	result = messageSplit("_LABEL__VALUE__MAXLENGTH_", placeHolders);
	assert(std.string.cmp(result[0], "_LABEL_") == 0);
	assert(std.string.cmp(result[1], "_VALUE_") == 0);
	assert(std.string.cmp(result[2], "_MAXLENGTH_") == 0);
	assert(std.string.cmp(messageBuild(result, replace), "ラベル値長さ") == 0);
}


/++
 + フォーム基底クラス
 +/
class Form
{
	/// s:idに指定する名前
	string name;

	/// 送信先のパス
	string action;

	/// GETにしたいときにtrueを設定
	bool get;

	/// enctypeをmultipart/form-dataにしたいときに指定
	bool multipart;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  name = s:idに指定する名前
	 +  action = 送信先のパス
	 +/
	this(string name, string action)
	{
		this.name = name;
		this.action = action;
	}

	/++
	 + フォームのデータをリクエストからロード
	 +
	 + Params:
	 +  request = リクエストデータ
	 +/
	void loadFromRequest(Request request)
	{
		convert_ = false;
		validate_ = true;

		foreach (field; fields_)
		{
			field.loadFromRequest(request);
			convert_ = convert_ || field.isConverted();
			validate_ = !(!validate_ || field.hasError());
		}
	}

	/++
	 + 凍結設定
	 +
	 + Params:
	 +  f = true:凍結 false:解凍
	 +/
	void frozen(bool f)
	{
		foreach (field; fields_)
			field.frozen = f;
	}

	/++
	 + SweetsにFormの出力処理をセット
	 +
	 + Params:
	 +  sweets = 対象のSweetsオブジェクト
	 +/
	void assignToSweets(Sweets sweets)
	{
		sweets.assign(name, &print);

		foreach (field; fields_)
			field.assignToSweets(sweets);
	}

	/++
	 + 変換が発生したか調べる
	 +
	 + Returns:
	 +  変換が発生していたらtrueを返す
	 +/
	const
	bool isConverted()
	{
		return convert_;
	}

	/++
	 + エラーが発生したか調べる
	 +
	 + Returns:
	 +  エラーが発生していたらtrueを返す
	 +/
	const
	bool hasError()
	{
		return !validate_;
	}

protected:

	/// 変換結果
	bool convert_;

	/// 入力チェック結果
	bool validate_;

	/// 保持しているフィールドのインデックス
	Field[string] fields_;

	/++
	 + フィールドを登録
	 +
	 + Fieldクラスのコンストラクタで自動的に呼ばれるので
	 + 手動で呼び出す必要はない
	 +/
	void registerIntoIndex(Field field)
	{
		fields_[field.name] = field;
	}

	/++
	 + HTMLタグとして出力
	 +
	 + Params:
	 +  element = 処理対象の要素
	 +  printer = 出力delegate
	 +/
	void print(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		// formタグが大文字とかの場合はそのまま出し、formタグ以外の場合は強制的にformタグとする
		if (std.string.icmp(element.name, "form") == 0)
		{
			printer("<");
			printer(element.name);
		}
		else
			printer("<form");

		// テンプレートにaction, method,enctypeの属性がかかれているかチェックする
		bool actionExists = false;
		bool methodExists = false;
		bool enctypeExists = false;

		// 全ての属性を出力
		foreach (k, v; element.attributes)
		{
			printer(" ");

			// action, method, enctypeはテンプレート側の設定を無視してプログラム側から指定する
			if (std.string.icmp(k, "action") == 0)
			{
				printer(k);
				printer("=\"");
				printer(escape(action));
				printer("\"");

				actionExists = true;
			}
			else if (std.string.icmp(k, "method") == 0)
			{
				printer(k);
				printer("=\"");
				printer(get ? "GET" : "POST");
				printer("\"");

				methodExists = true;
			}
			else if (std.string.icmp(k, "enctype") == 0)
			{
				if (multipart)
				{
					printer(k);
					printer("=\"multipart/form-data\"");
				}

				enctypeExists = true;
			}
			else
			{
				printer(k);
				printer("=\"");
				printer(v);
				printer("\"");
			}
		}


		// action, method, enctypeが未出力の場合は別途処理する

		if (!actionExists)
		{
			printer(" action=\"");
			printer(escape(action));
			printer("\"");
		}

		if (!methodExists)
		{
			printer(" method=\"");
			printer(get ? "GET" : "POST");
			printer("\"");
		}

		if (!enctypeExists && multipart)
			printer(" enctype=\"multipart/form-data\"");

		printer(">");

		// 中身を出力
		foreach (child ; children)
			child.print(printer);

		// 終了タグを出力
		element.printEndTag(printer);
	}
}

/++
 + 入力フィールドの基底クラス
 +/
class Field
{
	/// ラベルを表示するタグのIDに付ける接尾詞
	static immutable
	string LABEL_SUFFIX = "Label";

	/// 注釈をループするタグのIDに付ける接尾詞
	static immutable
	string ANNOTATIONS_SUFFIX = "Annotations";

	/// ループ中でカレントの注釈を出力するタグのIDに付ける接尾詞
	static immutable
	string ANNOTATION_SUFFIX = "Annotation";

	/// エラーをループするタグのIDに付ける接尾詞
	static immutable
	string ERRORS_SUFFIX = "Errors";

	/// ループ中でカレントのエラーを出力するタグのIDに付ける接尾詞
	static immutable
	string ERROR_SUFFIX = "Error";


	/// 凍結フラグ
	bool frozen;

	/// HTML上で使用する表示用文字列
	string label;

	/// リクエストで使用するキー文字列
	string name;

	/// 値
	string value;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  form = 親となるFormオブジェクト
	 +  name = リクエストで使用するキー文字列
	 +  label = HTML上で使用する表示用文字列
	 +/
	this(Form form, string name, string label)
	{
		this.name = name;
		this.label = label;

		// フォームのインデックスに自身を登録
		form.registerIntoIndex(this);
	}

	/++
	 + 変換が発生したか調べる
	 +
	 + Returns:
	 +  変換が発生していたらtrueを返す
	 +/
	const
	bool isConverted()
	{
		return convert_;
	}

	/++
	 + エラーが発生したか調べる
	 +
	 + Returns:
	 +  エラーが発生していたらtrueを返す
	 +/
	const
	bool hasError()
	{
		return !validate_;
	}

	/++
	 + フォームのデータをリクエストからロード
	 +
	 + Params:
	 +  request = リクエストデータ
	 +/
	void loadFromRequest(Request request)
	{
		convert_ = false;
		validate_ = true;
		annotations_.length = 0;
		errors_.length = 0;

		value = request.getParameter(name);

		foreach (converter; converters_)
		{
			if (converter.convert())
			{
				convert_ = true;
				string message = converter.message;
				if (message !is null && message.length > 0)
					annotations_ ~= message;
			}
		}

		foreach (validator; validators_)
		{
			if (!validator.validate())
			{
				validate_ = false;
				string message = validator.message;
				if (message !is null && message.length > 0)
					errors_ ~= message;
			}
		}
	}

	/++
	 + 変換で発生した注釈を取得
	 +
	 + Returns:
	 +  注釈
	 +/
	const
	const(string[]) annotations()
	{
		return annotations_;
	}

	/++
	 + 入力チェックで発生したエラーメッセージを取得
	 +
	 + Returns:
	 +  エラーメッセージ
	 +/
	const
	const(string[]) errors()
	{
		return errors_;
	}

	/++
	 + 注釈を追加
	 +
	 + 手動で変換処理をしたとき等に注釈を追加する
	 + なるべくConverterを実装したいところだが、
	 + 一カ所だけちょこっと処理を追加とかしたいときもあるので
	 +
	 + Params:
	 +  annotation = 注釈
	 +/
	void addAnnotation(string annotation)
	{
		annotations_ ~= annotation;
	}

	/++
	 + エラーを追加
	 +
	 + 手動でエラーチェックをしたとき等にエラーメッセージを追加する
	 + なるべくValidatorを実装したいところだが、
	 + 一カ所だけちょこっと処理を追加とかしたいときもあるので
	 +
	 + Params:
	 +  error = エラー
	 +/
	void addError(string error)
	{
		errors_ ~= error;
	}

	/++
	 + T型として取得
	 +
	 + Params:
	 +  defaultValue = 変換出来ない場合に返す値 指定しないとその型のinitプロパティ
	 +/
	const
	T getValueAs(T)(T defaultValue=T.init)
	{
		try
			return to!(T)(value);
		catch (Error e)
			return defaultValue;
	}

	/+
	 + getValueAsのbool型への変換の特殊化
	 +
	 + trueと判定する: 1, t, true
	 + falseと判定する: 0, f, false
	 + ※大文字小文字区別なし
	 +
	 + Params:
	 +  defaultValue = 変換出来ない場合に返す値
	 +/
	const
	T getValueAs(T:bool)(T defaultValue=T.init)
	{
		if (cmp(value, "1") == 0 || icmp(value, "t") == 0 || icmp(value, "true") == 0)
			return true;
		else if (cmp(value, "0") == 0 || icmp(value, "f") == 0 || icmp(value, "false") == 0)
			return false;
		else
			return defaultValue;
	}

	/++
	 + SweetsにFormの出力処理をセット
	 +
	 + Params:
	 +  sweets = 対象のSweetsオブジェクト
	 +/
	void assignToSweets(Sweets sweets)
	{
		sweets.assign(name, &print);
		sweets.assign(name ~ LABEL_SUFFIX, &printLabel);
		sweets.assign(name ~ ANNOTATIONS_SUFFIX, &printAnnotations);
		sweets.assign(name ~ ANNOTATION_SUFFIX, &printAnnotation);
		sweets.assign(name ~ ERRORS_SUFFIX, &printErrors);
		sweets.assign(name ~ ERROR_SUFFIX, &printError);
	}

protected:

	/// 変換結果
	bool convert_;

	/// 入力チェック結果
	bool validate_;

	/// 変換で発生した注釈
	string[] annotations_;

	/// 入力チェックで発生したエラーメッセージ
	string[] errors_;

	/// 変換処理
	Converter[] converters_;

	/// 入力チェック処理
	Validator[] validators_;

	/// ループ中の注釈
	string currentAnnotation_;

	/// ループ中のエラー
	string currentError_;

	/++
	 + 変換処理を追加
	 +
	 + Params:
	 +  converter = 変換処理
	 +/
	void addConverter(Converter converter)
	{
		converters_ ~= converter;
	}

	/++
	 + エラーチェック処理を追加
	 +
	 + Params:
	 +  validator = エラーチェック処理
	 +/
	void addValidator(Validator validator)
	{
		validators_ ~= validator;
	}

	/++
	 + 入力フィールドをHTMLタグとして出力
	 +
	 + Params:
	 +  element = 処理対象の要素
	 +  printer = 出力delegate
	 +/
	void print(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		if (frozen)
			printAtFrozen(printer, element, children);
		else
			printAtNormal(printer, element, children);
	}

	/++
	 + 通常状態を出力
	 +
	 + 各派生クラスで実装する
	 +
	 + Params:
	 +  element = 処理対象の要素
	 +  printer = 出力delegate
	 +/
	abstract
	void printAtNormal(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children);

	/++
	 + フリーズ状態を出力
	 +
	 + Params:
	 +  element = 処理対象の要素
	 +  printer = 出力delegate
	 +/
	void printAtFrozen(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		printer("<input type=\"hidden\" name=\"");
		printer(escape(name));
		printer("\" value=\"");
		printer(escape(value));
		printer("\"/>");
	}

	/++
	 + ラベルを出力
	 +
	 + Params:
	 +  element = 処理対象の要素
	 +  printer = 出力delegate
	 +/
	void printLabel(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		element.printStartTag(printer, false);
		printer(escape(label));
		element.printEndTag(printer);
	}

	/++
	 + 注釈をループして出力
	 +
	 + Params:
	 +  element = 処理対象の要素
	 +  printer = 出力delegate
	 +/
	void printAnnotations(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		const(string)* f = "s:foreach" in element.attributes;
		if (f !is null && std.string.cmp(*f, "together") == 0)
		{
			foreach (s; annotations_)
			{
				element.printStartTag(printer, false);
				currentAnnotation_ = s;
				foreach (child; children)
					child.print(printer);
				element.printEndTag(printer);
			}
		}
		else if (f !is null && std.string.cmp(*f, "insideonly") == 0)
		{
			foreach (s; annotations_)
			{
				currentAnnotation_ = s;
				foreach (child; children)
					child.print(printer);
			}
		}
		else
		{
			element.printStartTag(printer, false);
			foreach (s; annotations_)
			{
				currentAnnotation_ = s;
				foreach (child; children)
					child.print(printer);
			}
			element.printEndTag(printer);
		}
	}

	/++
	 + 注釈を1件出力
	 +
	 + Params:
	 +  element = 処理対象の要素
	 +  printer = 出力delegate
	 +/
	void printAnnotation(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		element.printStartTag(printer, false);
		printer(escape(currentAnnotation_));
		element.printEndTag(printer);
	}

	/++
	 + エラーをループして出力
	 +
	 + Params:
	 +  element = 処理対象の要素
	 +  printer = 出力delegate
	 +/
	void printErrors(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		const(string)* f = "s:foreach" in element.attributes;
		if (f !is null && std.string.cmp(*f, "together") == 0)
		{
			foreach (s; errors_)
			{
				element.printStartTag(printer, false);
				currentError_ = s;
				foreach (child; children)
					child.print(printer);
				element.printEndTag(printer);
			}
		}
		else if (f !is null && std.string.cmp(*f, "insideonly") == 0)
		{
			foreach (s; errors_)
			{
				currentError_ = s;
				foreach (child; children)
					child.print(printer);
			}
		}
		else
		{
			element.printStartTag(printer, false);
			foreach (s; errors_)
			{
				currentError_ = s;
				foreach (child; children)
					child.print(printer);
			}
			element.printEndTag(printer);
		}
	}

	/++
	 + エラーを1件出力
	 +
	 + Params:
	 +  element = 処理対象の要素
	 +  printer = 出力delegate
	 +/
	void printError(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		element.printStartTag(printer, false);
		printer(escape(currentError_));
		element.printEndTag(printer);
	}
}

/++
 + Submitボタンのクラス
 +/
class SubmitField : Field
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  form = 親となるFormオブジェクト
	 +  name = リクエストで使用するキー文字列
	 +  label = HTML上で使用する表示用文字列
	 +/
	this(Form form, string name, string label)
	{
		super(form, name, label);
	}

	/++
	 + ボタンが押されたか判定する
	 +
	 + Returns:
	 +  押されていればtrueを返す
	 +/
	bool isSubmitted()
	{
		return value !is null;
	}

protected:

	override
	void printAtNormal(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		printer("<input type=\"submit\" name=\"");
		printer(escape(name));
		printer("\" value=\"");
		printer(escape(label));
		printer("\"/>");
	}

	override
	void printAtFrozen(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		printAtNormal(printer, element, children);
	}
}

/++
 + Hiddenタグのクラス
 +/
class HiddenField : Field
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  form = 親となるFormオブジェクト
	 +  name = リクエストで使用するキー文字列
	 +/
	this(Form form, string name)
	{
		super(form, name, "");
	}

protected:

	override
	void printAtNormal(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		printAtFrozen(printer, element, children);
	}
}

/++
 + テキスト入力フィールドのクラス
 +
 + テキストボックスまたはテキストエリア
 +/
class TextInputField : Field
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  form = 親となるFormオブジェクト
	 +  name = リクエストで使用するキー文字列
	 +  label = HTML上で使用する表示用文字列
	 +/
	this(Form form, string name, string label)
	{
		super(form, name, label);
	}

protected:

	override
	void printAtNormal(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		if (std.string.icmp(element.name, "textarea") == 0)
		{
			printer("<");
			printer(element.name);

			bool nameExists = false;
			foreach (k, v; element.attributes)
			{
				printer(" ");

				if (std.string.icmp(k, "name") == 0)
				{
					printer(k);
					printer("=\"");
					printer(escape(name));
					printer("\"");

					nameExists = true;
				}
				else
				{
					printer(k);
					printer("=\"");
					printer(v);
					printer("\"");
				}
			}

			if (!nameExists)
			{
				printer(" name=\"");
				printer(escape(name));
				printer("\"");
			}

			printer(">");
	
			printer(escape(value));
			element.printEndTag(printer);
		}
		else
		{
			if (std.string.icmp(element.name, "input") == 0)
			{
				printer("<");
				printer(element.name);
			}
			else
				printer("<input");

			bool typeExists = false;
			bool nameExists = false;
			bool valueExists = false;
			foreach (k, v; element.attributes)
			{
				printer(" ");

				if (std.string.icmp(k, "type") == 0)
				{
					printer(k);

					if (std.string.icmp(v, "text") == 0)
					{
						printer("=\"");
						printer(v);
						printer("\"");
					}
					else
					{
						printer("=\"text\"");
					}

					typeExists = true;
				}
				else if (std.string.icmp(k, "name") == 0)
				{
					printer(k);
					printer("=\"");
					printer(escape(name));
					printer("\"");

					nameExists = true;
				}
				else if (std.string.icmp(k, "value") == 0)
				{
					printer(k);
					printer("=\"");
					printer(escape(value));
					printer("\"");

					valueExists = true;
				}
				else
				{
					printer(k);
					printer("=\"");
					printer(v);
					printer("\"");
				}
			}

			if (!typeExists)
			{
				printer(" type=\"text\"");
			}

			if (!nameExists)
			{
				printer(" name=\"");
				printer(escape(name));
				printer("\"");
			}

			if (!valueExists)
			{
				printer(" value=\"");
				printer(escape(value));
				printer("\"");
			}

			printer("/>");
		}
	}
}

/++
 + パスワードフィールドのクラス
 +/
class PasswordField : Field
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  form = 親となるFormオブジェクト
	 +  name = リクエストで使用するキー文字列
	 +  label = HTML上で使用する表示用文字列
	 +/
	this(Form form, string name, string label)
	{
		super(form, name, label);
	}

protected:

	override
	void printAtNormal(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		if (std.string.icmp(element.name, "input") == 0)
		{
			printer("<");
			printer(element.name);
		}
		else
			printer("<input");

		bool typeExists = false;
		bool nameExists = false;
		bool valueExists = false;
		foreach (k, v; element.attributes)
		{
			printer(" ");

			if (std.string.icmp(k, "type") == 0)
			{
				printer(k);

				if (std.string.icmp(v, "password") == 0)
				{
					printer("=\"");
					printer(v);
					printer("\"");
				}
				else
					printer("=\"password\"");

				typeExists = true;
			}
			else if (std.string.icmp(k, "name") == 0)
			{
				printer(k);
				printer("=\"");
				printer(escape(name));
				printer("\"");

				nameExists = true;
			}
			else if (std.string.icmp(k, "value") == 0)
			{
				printer(k);
				printer("=\"");
				printer(escape(value));
				printer("\"");

				valueExists = true;
			}
			else
			{
				printer(k);
				printer("=\"");
				printer(v);
				printer("\"");
			}
		}

		if (!typeExists)
		{
			printer(" type=\"text\"");
		}

		if (!nameExists)
		{
			printer(" name=\"");
			printer(escape(name));
			printer("\"");
		}

		if (!valueExists)
		{
			printer(" value=\"");
			printer(escape(value));
			printer("\"");
		}

		printer("/>");
	}
}

/++
 + チェックボックスクラス
 +/
class CheckboxField : Field
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  form = 親となるFormオブジェクト
	 +  name = リクエストで使用するキー文字列
	 +  label = HTML上で使用する表示用文字列
	 +  checkedValue = チェックされた時に送信される値 デフォルトは"1"
	 +/
	this(Form form, string name, string label, string checkedValue = null)
	{
		if (checkedValue is null || checkedValue.length == 0)
			checkedValue_ = "1";
		else
			checkedValue_ = checkedValue;

		super(form, name, label);
	}

	/++
	 + チェックされているか判定
	 +
	 + Params:
	 +  チェックされていればtrueを返す
	 +/
	bool isChecked()
	{
		return value !is null && std.string.cmp(value, checkedValue_) == 0;
	}

protected:

	/// チェックされた時に送信される値
	string checkedValue_;

	override
	void printAtNormal(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		if (std.string.icmp(element.name, "input") == 0)
		{
			printer("<");
			printer(element.name);
		}
		else
			printer("<input");

		bool typeExists = false;
		bool nameExists = false;
		bool valueExists = false;
		bool checkedExists = false;
		foreach (k, v; element.attributes)
		{
			printer(" ");

			if (std.string.icmp(k, "type") == 0)
			{
				printer(k);

				if (std.string.icmp(v, "checkbox") == 0)
				{
					printer("=\"");
					printer(v);
					printer("\"");
				}
				else
				{
					printer("=\"checkbox\"");
				}

				typeExists = true;
			}
			else if (std.string.icmp(k, "name") == 0)
			{
				printer(k);
				printer("=\"");
				printer(escape(name));
				printer("\"");

				nameExists = true;
			}
			else if (std.string.icmp(k, "value") == 0)
			{
				printer(k);
				printer("=\"");
				printer(escape(checkedValue_));
				printer("\"");

				valueExists = true;
			}
			else if (std.string.icmp(k, "checked") == 0)
			{
				if (isChecked())
				{
					printer(k);
					printer("=\"checked\"");
				}

				valueExists = true;
			}
			else
			{
				printer(k);
				printer("=\"");
				printer(v);
				printer("\"");
			}
		}

		if (!typeExists)
		{
			printer(" type=\"text\"");
		}

		if (!nameExists)
		{
			printer(" name=\"");
			printer(escape(name));
			printer("\"");
		}

		if (!valueExists)
		{
			printer(" value=\"");
			printer(escape(checkedValue_));
			printer("\"");
		}

		if (!checkedExists)
		{
			if (isChecked())
				printer(" checked=\"checked\"");
		}

		printer("/>");
	}
}

/++
 + 選択肢クラス
 +/
class SelectOption
{
	/// HTML上で使用する表示用文字列
	string label;

	/// 選択された時に送信される値
	string value;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  label = HTML上で使用する表示用文字列
	 +  value = 選択された時に送信される値
	 +/
	this(string label, string value)
	{
		this.label = label;
		this.value = value;
	}
}

/++
 + 複数選択肢から1つを選択するフィールドのクラス
 +
 + セレクトボックスまたはラジオボタン
 +/
class SingleSelectField : Field
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  form = 親となるFormオブジェクト
	 +  name = リクエストで使用するキー文字列
	 +  label = HTML上で使用する表示用文字列
	 +  options = 選択肢の配列
	 +/
	this(Form form, string name, string label, SelectOption[] options)
	{
		super(form, name, label);

		options_ = options;
		memoryPrinter_ = new MemoryPrinter;
	}

	override
	void loadFromRequest(Request request)
	{
		convert_ = false;
		validate_ = true;
		annotations_.length = 0;
		errors_.length = 0;

		if (options_ is null)
			value = null;
		else
		{
			string v = request.getParameter(name);

			foreach (option; options_)
			{
				if (std.string.cmp(v, option.value) == 0)
				{
					value = v;
					break;
				}
			}
		}

		foreach (converter; converters_)
		{
			if (converter.convert())
			{
				convert_ = true;
				string message = converter.message;
				if (message !is null && message.length > 0)
					annotations_ ~= message;
			}
		}

		foreach (validator; validators_)
		{
			if (!validator.validate())
			{
				validate_ = false;
				string message = validator.message;
				if (message !is null && message.length > 0)
					errors_ ~= message;
			}
		}
	}

	/++
	 + 選択肢を取得
	 +
	 + Returns:
	 +  選択肢
	 +/
	const(SelectOption[]) options()
	{
		return options_;
	}

	/++
	 + 選択肢を設定
	 +
	 + Params:
	 +  options = 選択肢
	 +/
	void options(SelectOption[] options)
	{
		options_ = options;

		if (value !is null)
		{
			if (options_ !is null)
			{
				foreach (option; options_)
				{
					if (std.string.cmp(value, option.value) == 0)
						return;
				}
			}
			value = null;
		}
	}

protected:

	/// 選択肢
	SelectOption[] options_;

	/// 区切り文字を解析するときに使用
	MemoryPrinter memoryPrinter_;

	override
	void printAtNormal(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		if (std.string.icmp(element.name, "select") == 0)
		{
			printer("<");
			printer(element.name);

			bool nameExists = false;
			foreach (k, v; element.attributes)
			{
				printer(" ");

				if (std.string.icmp(k, "name") == 0)
				{
					printer(k);
					printer("=\"");
					printer(escape(name));
					printer("\"");

					nameExists = true;
				}
				else
				{
					printer(k);
					printer("=\"");
					printer(v);
					printer("\"");
				}
			}

			if (!nameExists)
			{
				printer(" name=\"");
				printer(escape(name));
				printer("\"");
			}

			printer(">");

			if (options_ !is null)
			{
				foreach (option; options_)
				{
					printer("<option value=\"");
					printer(escape(option.value));
					if (std.string.cmp(value, option.value) == 0)
						printer("\" selected=\"selected\">");
					else
						printer("\">");
					printer(escape(option.label));
					printer("</option>");
				}
			}

			element.printEndTag(printer);
		}
		else
		{
			// 子タグを解析して区切り文字とする
			memoryPrinter_.clear();
			foreach (child; children)
				child.print(&memoryPrinter_.print);
			const(char)[] delimiter = memoryPrinter_.buffer;

			if (options_ !is null)
			{
				bool first = true;
				foreach (option; options_)
				{
					if (!first)
						printer(delimiter);

					printer("<label><input type=\"radio\" name=\"");
					printer(escape(name));
					printer("\" value=\"");
					printer(escape(option.value));
					if (std.string.cmp(value, option.value) == 0)
						printer("\" checked=\"checked\"/>");
					else
						printer("\"/>");
					printer(escape(option.label));
					printer("</label>");

					first = false;
				}
			}
		}
	}
}

/++
 + 複数選択肢から複数を選択するフィールドのクラス
 +
 + セレクトボックスまたはチェックボックス
 +/
class MultipleSelectField : Field
{
	/// 選択された値の配列
	string[] values;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  form = 親となるFormオブジェクト
	 +  name = リクエストで使用するキー文字列
	 +  label = HTML上で使用する表示用文字列
	 +  options = 選択肢の配列
	 +/
	this(Form form, string name, string label, SelectOption[] options)
	{
		super(form, name, label);

		options_ = options;
		memoryPrinter_ = new MemoryPrinter;
	}

	override
	void loadFromRequest(Request request)
	{
		convert_ = false;
		validate_ = true;
		annotations_.length = 0;
		errors_.length = 0;

		values = [];
		if (options_ !is null)
		{
			const(char[][]) sourceValues = request.getParametersDirect(name);
			foreach (option; options_)
			{
				foreach (value; sourceValues)
				{
					if (std.string.cmp(value, option.value) == 0)
					{
						values ~= option.value;
						break;
					}
				}
			}
		}

		foreach (converter; converters_)
		{
			if (converter.convert())
			{
				convert_ = true;
				string message = converter.message;
				if (message !is null && message.length > 0)
					annotations_ ~= message;
			}
		}

		foreach (validator; validators_)
		{
			if (!validator.validate())
			{
				validate_ = false;
				string message = validator.message;
				if (message !is null && message.length > 0)
					errors_ ~= message;
			}
		}
	}

	/++
	 + 選択肢を取得
	 +
	 + Returns:
	 +  選択肢
	 +/
	const(SelectOption[]) options()
	{
		return options_;
	}

	/++
	 + 選択肢を設定
	 +
	 + Params:
	 +  options = 選択肢
	 +/
	void options(SelectOption[] options)
	{
		options_ = options;

		string[] newValues = [];
		if (options_ !is null)
		{
			foreach (option; options_)
			{
				foreach (value; values)
				{
					if (std.string.cmp(option.value, value) == 0)
					{
						newValues ~= value;
						break;
					}
				}
			}
		}
		values = newValues;
	}

protected:

	/// 選択肢
	SelectOption[] options_;

	/// 区切り文字を解析するときに使用
	MemoryPrinter memoryPrinter_;

	override
	void printAtNormal(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		// 選択済みの値を取得
		uint[string] selected;
		foreach (value; values)
			selected[value] = 1;

		if (std.string.icmp(element.name, "select") == 0)
		{
			printer("<");
			printer(element.name);

			bool nameExists = false;
			bool multipleExists = false;
			foreach (k, v; element.attributes)
			{
				printer(" ");

				if (std.string.icmp(k, "name") == 0)
				{
					printer(k);
					printer("=\"");
					printer(escape(name));
					printer("\"");

					nameExists = true;
				}
				else if (std.string.icmp(k, "multiple") == 0)
				{
					printer(k);
					if (std.string.icmp(v, "multiple") == 0)
					{
						printer("=\"");
						printer(v);
						printer("\"");
					}
					else
						printer("=\"multiple\"");

					multipleExists = true;
				}
				else
				{
					printer(k);
					printer("=\"");
					printer(v);
					printer("\"");
				}
			}

			if (!nameExists)
			{
				printer(" name=\"");
				printer(escape(name));
				printer("\"");
			}

			if (!multipleExists)
				printer(" multiple=\"multiple\"");

			printer(">");

			if (options_ !is null)
			{
				foreach (option; options_)
				{
					printer("<option value=\"");
					printer(escape(option.value));
					if ((option.value in selected) is null)
						printer("\">");
					else
						printer("\" selected=\"selected\">");
					printer(escape(option.label));
					printer("</option>");
				}
			}

			element.printEndTag(printer);
		}
		else
		{
			// 子タグを解析して区切り文字とする
			memoryPrinter_.clear();
			foreach (child; children)
				child.print(&memoryPrinter_.print);
			const(char)[] delimiter = memoryPrinter_.buffer;

			if (options_ !is null)
			{
				bool first = true;
				foreach (option; options_)
				{
					if (!first)
						printer(delimiter);

					printer("<label><input type=\"checkbox\" name=\"");
					printer(escape(name));
					printer("\" value=\"");
					printer(escape(option.value));
					if ((option.value in selected) is null)
						printer("\">");
					else
						printer("\" checked=\"checked\">");
					printer(escape(option.label));
					printer("</label>");

					first = false;
				}
			}
		}
	}

	override
	void printAtFrozen(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		string before =  "<input type=\"hidden\" name=\"" ~ escape(name) ~ "\" value=\"";
		foreach (v; values)
		{
			printer(before);
			printer(escape(value));
			printer("\"/>");
		}
	}
}

/++
 + 日時のセレクトボックス
 +/
class DateTimeField : Field
{
	// フィールドが未選択の場合の値
	/// blank = trueの場合に、フィールドが範囲外の時はこの数値にセットされる
	static immutable
	byte OPTION_BLANK_VALUE = -1;

	@property
	{
		/// 開始年
		short startYear()
		{
			return startYear_;
		}

		void startYear(short y)
		{
			if (y < 0 || 9999 < y)
				startYear_ = 0;
			else
				startYear_ = y;
		}


		/// 終了年
		short endYear()
		{
			return endYear_;
		}

		void endYear(short y)
		{
			if (y < 0 || 9999 < y)
				endYear_ = 9999;
			else
				endYear_ = y;
		}


		/// 年
		short year()
		{
			return year_;
		}

		void year(short y)
		{
			if (y < startYear || endYear < y)
			{
				if (blank_)
					year_ = OPTION_BLANK_VALUE;
				else
					year_ = startYear;
			}
			else
				year_ = y;
		}


		/// 月
		byte month()
		{
			return month_;
		}

		void month(byte m)
		{
			if (m < 1 || 12 < m)
			{
				if (blank_)
					month_ = OPTION_BLANK_VALUE;
				else
					month_ = 1;
			}
			else
				month_ = m;
		}


		/// 日
		byte day()
		{
			return day_;
		}

		void day(byte d)
		{
			if (d < 1 || 31 < d)
			{
				if (blank_)
					day_ = OPTION_BLANK_VALUE;
				else
					day_ = 1;
			}
			else
				day_ = d;
		}


		/// 時
		byte hour()
		{
			return hour_;
		}

		void hour(byte h)
		{
			if (h < 0 || 23 < h)
			{
				if (blank_)
					hour_ = OPTION_BLANK_VALUE;
				else
					hour_ = 0;
			}
			else
				hour_ = h;
		}


		/// 分
		byte minute()
		{
			return minute_;
		}

		void minute(byte m)
		{
			if (m < 0 || 59 < m)
			{
				if (blank_)
					minute_ = OPTION_BLANK_VALUE;
				else
					minute_ = 0;
			}
			else
				minute_ = m;
		}


		/// 秒
		byte second()
		{
			return second_;
		}

		void second(byte s)
		{
			if (s < 0 || 59 < s)
			{
				if (blank_)
					second_ = OPTION_BLANK_VALUE;
				else
					second_ = 0;
			}
			else
				second_ = s;
		}


		/// 未選択許可
		bool blank()
		{
			return blank_;
		}

		void blank(bool b)
		{
			blank_ = b;

			// 未選択を許可する/しないでフィールドの値を調整する
			year = year_;
			month = month_;
			day = day_;
			hour = hour_;
			minute = minute_;
			second = second_;
		}
	}


	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  form = 親となるFormオブジェクト
	 +  name = リクエストで使用するキー文字列
	 +  label = HTML上で使用する表示用文字列
	 +  startYear = 年の開始 (0 ～ 9999)
	 +  endYear = 年の終了 (0 ～ 9999)
	 +  blank = 未選択可能にする場合はtrueを指定する
	 +/
	this(Form form, string name, string label, short startYear, short endYear, bool blank = false)
	{
		super(form, name, label);

		// 渡されたキー文字列に、データ毎の接尾辞を追加
		for (uint i; i < RealField.LENGTH; ++i)
			names_[i] = name ~ REAL_FIELD_SUFFIXES[i];

		this.startYear = startYear;
		this.endYear = endYear;
		this.blank = blank;
	}

	override
	void loadFromRequest(Request request)
	{
		convert_ = false;
		validate_ = true;
		annotations_.length = 0;
		errors_.length = 0;

		year = request.getParameterAs!(short)(names_[RealField.YEAR]);
		month = request.getParameterAs!(byte)(names_[RealField.MONTH]);
		day = request.getParameterAs!(byte)(names_[RealField.DAY]);
		hour = request.getParameterAs!(byte)(names_[RealField.HOUR]);
		minute = request.getParameterAs!(byte)(names_[RealField.MINUTE]);
		second = request.getParameterAs!(byte)(names_[RealField.SECOND]);

		foreach (converter; converters_)
		{
			if (converter.convert())
			{
				convert_ = true;
				string message = converter.message;
				if (message !is null && message.length > 0)
					annotations_ ~= message;
			}
		}

		foreach (validator; validators_)
		{
			if (!validator.validate())
			{
				validate_ = false;
				string message = validator.message;
				if (message !is null && message.length > 0)
					errors_ ~= message;
			}
		}
	}

protected:

	// 日付けセレクトボックス生成専用オプション
	struct DateTimeSelectOption
	{
		// ラベル
		string label;

		// 値
		byte value;

		// 値を文字列化したキャッシュ
		string valueString;
	}

	// オプションのデータテーブル
	static immutable
	DateTimeSelectOption[60] OPTIONS =
	[
		{"00",  0,  "0"},
		{"01",  1,  "1"},
		{"02",  2,  "2"},
		{"03",  3,  "3"},
		{"04",  4,  "4"},
		{"05",  5,  "5"},
		{"06",  6,  "6"},
		{"07",  7,  "7"},
		{"08",  8,  "8"},
		{"09",  9,  "9"},
		{"10", 10, "10"},
		{"11", 11, "11"},
		{"12", 12, "12"},
		{"13", 13, "13"},
		{"14", 14, "14"},
		{"15", 15, "15"},
		{"16", 16, "16"},
		{"17", 17, "17"},
		{"18", 18, "18"},
		{"19", 19, "19"},
		{"20", 20, "20"},
		{"21", 21, "21"},
		{"22", 22, "22"},
		{"23", 23, "23"},
		{"24", 24, "24"},
		{"25", 25, "25"},
		{"26", 26, "26"},
		{"27", 27, "27"},
		{"28", 28, "28"},
		{"29", 29, "29"},
		{"30", 30, "30"},
		{"31", 31, "31"},
		{"32", 32, "32"},
		{"33", 33, "33"},
		{"34", 34, "34"},
		{"35", 35, "35"},
		{"36", 36, "36"},
		{"37", 37, "37"},
		{"38", 38, "38"},
		{"39", 39, "39"},
		{"40", 40, "30"},
		{"41", 41, "41"},
		{"42", 42, "42"},
		{"43", 43, "43"},
		{"44", 44, "44"},
		{"45", 45, "45"},
		{"46", 46, "46"},
		{"47", 47, "47"},
		{"48", 48, "48"},
		{"49", 49, "49"},
		{"50", 50, "50"},
		{"51", 51, "51"},
		{"52", 52, "52"},
		{"53", 53, "53"},
		{"54", 54, "54"},
		{"55", 55, "55"},
		{"56", 56, "56"},
		{"57", 57, "57"},
		{"58", 58, "58"},
		{"59", 59, "59"},
	];

	// 各パラメータ用のインデックス
	enum RealField
	{
		YEAR,   // 年
		MONTH,  // 月
		DAY,    // 日
		HOUR,   // 時
		MINUTE, // 分
		SECOND, // 秒
		LENGTH, // 総数
	}

	// リクエストから各データを取得する時のキー文字列に付ける接尾辞
	static immutable
	string[RealField.LENGTH] REAL_FIELD_SUFFIXES =
	[
		"Year",
		"Month",
		"Day",
		"Hour",
		"Minute",
		"Second",
	];

	// 開始年
	short startYear_;

	// 終了年
	short endYear_;

	// 年
	short year_;

	// 月
	byte month_;

	// 日
	byte day_;

	// 時
	byte hour_;

	// 分
	byte minute_;

	// 秒
	byte second_;

	// 未選択許可フラグ
	bool blank_;

	// リクエストから各データを取得する時のキー文字列
	string[RealField.LENGTH] names_;

	override
	void printAtNormal(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		string format;
		const(string)* f = "s:format" in element.attributes;
		if (f is null)
			format = "%Y-%m-%d %H:%M:%S";
		else
			format = *f;

		uint printStart = 0;
		uint printEnd = 0;
		for (; printEnd < format.length; ++printEnd)
		{
			if (format[printEnd] == '%')
			{
				++printEnd;
				if (printEnd == format.length)
				{
					printer(format[printStart .. $]);
					return;
				}
				else
				{
					if (printStart + 1 < printEnd)
						printer(format[printStart .. printEnd - 1]);
					printStart = printEnd + 1;

					switch (format[printEnd])
					{
					case '%':
						printer("%");
						break;
					case 'n':
						printer("\n");
						break;
					case 't':
						printer("\t");
						break;
					case 'Y':
						printYearSelectbox(printer);
						break;
					case 'm':
						printSelectbox(printer, names_[RealField.MONTH], month_, OPTIONS[1..13]);
						break;
					case 'd':
						printSelectbox(printer, names_[RealField.DAY], day_, OPTIONS[1..32]);
						break;
					case 'H':
						printSelectbox(printer, names_[RealField.HOUR], hour_, OPTIONS[0..24]);
						break;
					case 'M':
						printSelectbox(printer, names_[RealField.MINUTE], minute_, OPTIONS);
						break;
					case 'S':
						printSelectbox(printer, names_[RealField.SECOND], second_, OPTIONS);
						break;
					default:
						printer("%" ~ format[printEnd]);
						break;
					}
				}
			}
		}

		printer(format[printStart .. $]);
	}

	override
	void printAtFrozen(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		if (year_ != OPTION_BLANK_VALUE)
		{
			printer("<input type=\"hidden\" name=\"");
			printer(escape(names_[RealField.YEAR]));
			printer("\" value=\"");
			printer(std.conv.to!string(year_));
			printer("\"/>");
		}

		if (month_ != OPTION_BLANK_VALUE)
		{
			printer("<input type=\"hidden\" name=\"");
			printer(escape(names_[RealField.MONTH]));
			printer("\" value=\"");
			printer(OPTIONS[month_].valueString);
			printer("\"/>");
		}

		if (day_ != OPTION_BLANK_VALUE)
		{
			printer("<input type=\"hidden\" name=\"");
			printer(escape(names_[RealField.DAY]));
			printer("\" value=\"");
			printer(OPTIONS[day_].valueString);
			printer("\"/>");
		}

		if (hour_ != OPTION_BLANK_VALUE)
		{
			printer("<input type=\"hidden\" name=\"");
			printer(escape(names_[RealField.HOUR]));
			printer("\" value=\"");
			printer(OPTIONS[hour_].valueString);
			printer("\"/>");
		}

		if (minute_ != OPTION_BLANK_VALUE)
		{
			printer("<input type=\"hidden\" name=\"");
			printer(escape(names_[RealField.MINUTE]));
			printer("\" value=\"");
			printer(OPTIONS[minute_].valueString);
			printer("\"/>");
		}

		if (second_ != OPTION_BLANK_VALUE)
		{
			printer("<input type=\"hidden\" name=\"");
			printer(escape(names_[RealField.SECOND]));
			printer("\" value=\"");
			printer(OPTIONS[second_].valueString);
			printer("\"/>");
		}
	}

	/+
	 + 年のセレクトボックスを表示
	 +
	 + Params:
	 +  printer = SweetsPrinter
	 +/
	void printYearSelectbox(SweetsPrinter printer)
	{
		printer("<select name=\"");
		printer(escape(names_[RealField.YEAR]));
		printer("\">");

		if (blank_)
			printer("<option value=\"" ~ std.conv.to!string(OPTION_BLANK_VALUE) ~ "\"></option>");

		for (short i = startYear_; i <= endYear_; ++i)
		{
			string value = std.conv.to!string(i);
	
			printer("<option value=\"");
			printer(value);
			if (year_ == i)
				printer("\" selected=\"selected\">");
			else
				printer("\">");

			printer(value);

			printer("</option>");
		}

		printer("</select>");
	}

	/+
	 + セレクトボックスを表示
	 +
	 + Params:
	 +  printer = SweetsPrinter
	 +  name = 名称
	 +  value = 値
	 +  options = 選択肢
	 +/
	void printSelectbox(SweetsPrinter printer, string name, byte value, const(DateTimeSelectOption[]) options)
	{
		printer("<select name=\"");
		printer(escape(name));
		printer("\">");

		if (blank_)
			printer("<option value=\"" ~ std.conv.to!string(OPTION_BLANK_VALUE) ~ "\"></option>");

		for (uint i; i < options.length; ++i)
		{
			printer("<option value=\"");
			printer(options[i].valueString);
			if (value == options[i].value)
				printer("\" selected=\"selected\">");
			else
				printer("\">");
			printer(options[i].label);
			printer("</option>");
		}

		printer("</select>");
	}
}

/++
 + ファイルアップロード入力フィールドのクラス
 +/
class FileUploadField : Field
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  form = 親となるFormオブジェクト
	 +  name = リクエストで使用するキー文字列
	 +  label = HTML上で使用する表示用文字列
	 +/
	this(Form form, string name, string label)
	{
		super(form, name, label);

		// ファイルアップロード用にmultipart/form-dataにする
		form.multipart = true;
	}

	/++
	 + アップロードされたファイルを取得
	 +
	 + Returns:
	 +  アップロードされたファイル
	 +/
	UploadedFile uploadedFile()
	{
		return uploadedFile_;
	}

	override
	void loadFromRequest(Request request)
	{
		convert_ = false;
		validate_ = true;
		annotations_.length = 0;
		errors_.length = 0;
		uploadedFile_ = null;

		uploadedFile_ = request.getUploadedFile(name);
		value = uploadedFile_ is null ? null : uploadedFile_.name;

		foreach (converter; converters_)
		{
			if (converter.convert())
			{
				convert_ = true;
				string message = converter.message;
				if (message !is null && message.length > 0)
					annotations_ ~= message;
			}
		}

		foreach (validator; validators_)
		{
			if (!validator.validate())
			{
				validate_ = false;
				string message = validator.message;
				if (message !is null && message.length > 0)
					errors_ ~= message;
			}
		}
	}

protected:

	/// アップロードされたファイル
	UploadedFile uploadedFile_;

	override
	void printAtNormal(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		if (std.string.icmp(element.name, "input") == 0)
		{
			printer("<");
			printer(element.name);
		}
		else
			printer("<input");

		bool typeExists = false;
		bool nameExists = false;
		foreach (k, v; element.attributes)
		{
			printer(" ");

			if (std.string.icmp(k, "type") == 0)
			{
				printer(k);

				if (std.string.icmp(v, "file") == 0)
				{
					printer("=\"");
					printer(v);
					printer("\"");
				}
				else
				{
					printer("=\"file\"");
				}

				typeExists = true;
			}
			else if (std.string.icmp(k, "name") == 0)
			{
				printer(k);
				printer("=\"");
				printer(escape(name));
				printer("\"");

				nameExists = true;
			}
			else
			{
				printer(k);
				printer("=\"");
				printer(v);
				printer("\"");
			}
		}

		if (!typeExists)
		{
			printer(" type=\"file\"");
		}

		if (!nameExists)
		{
			printer(" name=\"");
			printer(escape(name));
			printer("\"");
		}

		printer("/>");
	}

	override
	void printAtFrozen(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		// fileは毎回渡さないとなのでfreezeできない
		// 確認画面を挟むときはセッションかなんかでアップロードしたファイルを保持する
		printAtNormal(printer, element, children);
	}
}

/++
 + フォームの入力値を変換
 +/
class Converter
{
	/// ラベルに置き換わるプレースホルダ
	static immutable
	string PLACE_HOLDER_LABEL = "_LABEL_";

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象のフィールド
	 +  message = 変換時のメッセージ
	 +/
	this(Field field, string message)
	{
		field_ = field;
		field_.addConverter(this);

		// TODO: phobosのバグ対応で一回別の変数に代入する
		string label = PLACE_HOLDER_LABEL;
		messageParts_ ~= std.array.replace(message, label, field.label);
	}

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象のフィールド
	 +  message = 変換時のメッセージ
	 +  placeHolders = プレースホルダの配列
	 +/
	this(Field field, string message, string[] placeHolders)
	{
		field_ = field;
		field_.addConverter(this);

		// TODO: phobosのバグ対応で一回別の変数に代入する
		string label = PLACE_HOLDER_LABEL;
		message = std.array.replace(message, label, field.label);
		messageParts_ = messageSplit(message, placeHolders);
	}

	/++
	 + 変換処理を行った際のメッセージを取得
	 +
	 + Returns:
	 +  メッセージ
	 +/
	string message()
	{
		return messageBuild(messageParts_, getPlaceHolderReplaces());
	}

	/++
	 + 変換処理
	 +
	 + Returns:
	 +  変換処理が行われた場合にtrueを返す
	 +/
	abstract
	bool convert();

protected:

	/// 対象の入力フィールド
	Field field_;

	/// 変換時のメッセージを分割したもの
	string[] messageParts_;

	/++
	 + プレースホルダに投入する値を取得
	 +
	 + 継承先でプレースホルダに設定する値を返すよう実装する
	 +/
	string[string] getPlaceHolderReplaces()
	{
		return null;
	}
}

/++
 + 0以上の整数に変換
 +
 + 受け取った値が、整数でない場合または負の整数だった場合に0に変換する
 +/
class IndexConverter : Converter
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  message = 変換時のメッセージ
	 +/
	this(Field field, string message)
	{
		super(field, message);
	}

	override
	bool convert()
	{
		if (field_.value is null || field_.value.length == 0)
		{
			field_.value = "0";
			return true;
		}

		try
			std.conv.to!(uint)(field_.value);
		catch (Error e)
		{
			field_.value = "0";
			return true;
		}

		return false;
	}
}

/++
 + 全角数字を半角数字に変換
 +
 + 全角数字以外の文字には何もしない
 +/
class HalfWidthNumeralConverter : Converter
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  message = 変換時のメッセージ
	 +/
	this(Field field, string message)
	{
		super(field, message);
	}

	override
	bool convert()
	{
		if (field_.value is null || field_.value.length == 0)
			return false;

		immutable(char)* iterator = field_.value.ptr;
		immutable(char)* end = iterator + field_.value.length - 2;

		// 変換対象の文字が何個あるかカウントする
		uint count = 0;
		while (iterator < end)
		{
			if ((*iterator & 0b01000000) == 0)
				++ iterator;
			else if ((*iterator & 0b11100000) == 0b11000000)
				iterator += 2;
			else if ((*iterator & 0b11110000) == 0b11100000)
			{
				if (iterator[0] == 0xEF && iterator[1] == 0xBC)
				{
					ubyte number = cast(ubyte)(iterator[2] - 0x90);
					if (number < 10)
						++count;
				}

				iterator += 3;
			}
			else if ((*iterator & 0b11111000) == 0b11110000)
				iterator += 4;
			else
				++ iterator;
		}

		if (count == 0)
			return false;

		char[] buffer = new char[field_.value.length - count * 2];
		char* write = buffer.ptr;

		// 文字を変換しつつ新規のバッファにコピー
		iterator = field_.value.ptr;
		while (iterator < end)
		{
			if ((*iterator & 0b01000000) == 0)
				*write++ = *iterator++;
			else if ((*iterator & 0b11100000) == 0b11000000)
			{
				*write++ = *iterator++;
				*write++ = *iterator++;
			}
			else if ((*iterator & 0b11110000) == 0b11100000)
			{
				if (iterator[0] == 0xEF && iterator[1] == 0xBC)
				{
					ubyte number = cast(ubyte)(iterator[2] - 0x90);
					if (number < 10)
					{
						*write++ = cast(char)(0x30 + number);
						iterator += 3;
						continue;
					}
				}

				*write++ = *iterator++;
				*write++ = *iterator++;
				*write++ = *iterator++;
			}
			else if ((*iterator & 0b11111000) == 0b11110000)
			{
				*write++ = *iterator++;
				*write++ = *iterator++;
				*write++ = *iterator++;
				*write++ = *iterator++;
			}
			else
				*write++ = *iterator++;
		}

		// 変換結果をセット
		field_.value = cast(string)buffer;

		return true;
	}
	unittest
	{
		auto form = new Form("", "");
		auto field = new TextInputField(form, "test", "テスト");
		auto converter = new HalfWidthNumeralConverter(field, "");

		field.value = null;
		converter.convert();
		assert(field.value is null);

		field.value = "";
		converter.convert();
		assert(std.string.cmp(field.value, "") == 0);

		field.value = "ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｖｗｘｙｚ";
		converter.convert();
		assert(std.string.cmp(field.value, "ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｖｗｘｙｚ") == 0);

		field.value = "ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ";
		converter.convert();
		assert(std.string.cmp(field.value, "ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ") == 0);

		field.value = "０１２３４５６７８９";
		converter.convert();
		assert(std.string.cmp(field.value, "0123456789") == 0);

		field.value = "あかさたなはまやらわ";
		converter.convert();
		assert(std.string.cmp(field.value, "あかさたなはまやらわ") == 0);
	}
}

/++
 + 全角アルファベットを半角アルファベットに変換
 +
 + 全角アルファベット以外の文字には何もしない
 +/
class HalfWidthAlphabetConverter : Converter
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  message = 変換時のメッセージ
	 +/
	this(Field field, string message)
	{
		super(field, message);
	}

	override
	bool convert()
	{
		if (field_.value is null || field_.value.length == 0)
			return false;

		immutable(char)* iterator = field_.value.ptr;
		immutable(char)* end = iterator + field_.value.length - 2;

		// 変換対象の文字が何個あるかカウントする
		uint count = 0;
		while (iterator < end)
		{
			if ((*iterator & 0b01000000) == 0)
				++ iterator;
			else if ((*iterator & 0b11100000) == 0b11000000)
				iterator += 2;
			else if ((*iterator & 0b11110000) == 0b11100000)
			{
				if (iterator[0] == 0xEF)
				{
					if (iterator[1] == 0xBC)
					{
						ubyte number = cast(ubyte)(iterator[2] - 0xA1);
						if (number < 26)
							++ count;
					}
					else if (iterator[1] == 0xBD)
					{
						ubyte number = cast(ubyte)(iterator[2] - 0x81);
						if (number < 26)
							++ count;
					}
				}

				iterator += 3;
			}
			else if ((*iterator & 0b11111000) == 0b11110000)
				iterator += 4;
			else
				++ iterator;
		}

		if (count == 0)
			return false;

		char[] buffer = new char[field_.value.length - count * 2];
		char* write = buffer.ptr;

		// 文字を変換しつつ新規のバッファにコピー
		iterator = field_.value.ptr;
		while (iterator < end)
		{
			if ((*iterator & 0b01000000) == 0)
				*write++ = *iterator++;
			else if ((*iterator & 0b11100000) == 0b11000000)
			{
				*write++ = *iterator++;
				*write++ = *iterator++;
			}
			else if ((*iterator & 0b11110000) == 0b11100000)
			{
				if (iterator[0] == 0xEF)
				{
					if (iterator[1] == 0xBC)
					{
						ubyte number = cast(ubyte)(iterator[2] - 0xA1);
						if (number < 26)
						{
							*write++ = cast(char)(0x41 + number);
							iterator += 3;
							continue;
						}
					}
					else if (iterator[1] == 0xBD)
					{
						ubyte number = cast(ubyte)(iterator[2] - 0x81);
						if (number < 26)
						{
							*write++ = cast(char)(0x61 + number);
							iterator += 3;
							continue;
						}
					}
				}

				*write++ = *iterator++;
				*write++ = *iterator++;
				*write++ = *iterator++;
			}
			else if ((*iterator & 0b11111000) == 0b11110000)
			{
				*write++ = *iterator++;
				*write++ = *iterator++;
				*write++ = *iterator++;
				*write++ = *iterator++;
			}
			else
				*write++ = *iterator++;
		}

		// 変換結果をセット
		field_.value = cast(string)buffer;

		return true;
	}
	unittest
	{
		auto form = new Form("", "");
		auto field = new TextInputField(form, "test", "テスト");
		auto converter = new HalfWidthAlphabetConverter(field, "");

		field.value = null;
		converter.convert();
		assert(field.value is null);

		field.value = "";
		converter.convert();
		assert(std.string.cmp(field.value, "") == 0);

		field.value = "ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｖｗｘｙｚ";
		converter.convert();
		assert(std.string.cmp(field.value, "abcdefghijklmnopqrstuvwxyz") == 0);

		field.value = "ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ";
		converter.convert();
		assert(std.string.cmp(field.value, "ABCDEFGHIJKLMNOPQRSTUVWXYZ") == 0);

		field.value = "０１２３４５６７８９";
		converter.convert();
		assert(std.string.cmp(field.value, "０１２３４５６７８９") == 0);

		field.value = "あかさたなはまやらわ";
		converter.convert();
		assert(std.string.cmp(field.value, "あかさたなはまやらわ") == 0);
	}
}

/++
 + 入力フィールドに対して入力値の検証を行う
 +/
class Validator
{
	/// ラベルに置き換わるプレースホルダ
	static immutable
	string PLACE_HOLDER_LABEL = "_LABEL_";

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象のフィールド
	 +  message = 変換時のメッセージ
	 +/
	this(Field field, string message)
	{
		field_ = field;
		field_.addValidator(this);

		string label = PLACE_HOLDER_LABEL;
		messageParts_ ~= std.array.replace(message, label, field.label);
	}

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象のフィールド
	 +  message = 変換時のメッセージ
	 +  placeHolders = プレースホルダの配列
	 +/
	this(Field field, string message, string[] placeHolders)
	{
		field_ = field;
		field_.addValidator(this);

		string label = PLACE_HOLDER_LABEL;
		message = std.array.replace(message, label, field.label);
		messageParts_ = messageSplit(message, placeHolders);
	}

	/++
	 + 変換処理を行った際のメッセージを取得
	 +
	 + Returns:
	 +  メッセージ
	 +/
	string message()
	{
		return messageBuild(messageParts_, getPlaceHolderReplaces());
	}

	/++
	 + 入力チェック
	 +
	 + Returns:
	 +  正常と判断された場合にtrueを返す
	 +/
	abstract
	bool validate();

protected:

	/// 対象の入力フィールド
	Field field_;

	/// 変換時のメッセージを分割したもの
	string[] messageParts_;

	/++
	 + プレースホルダに投入する値を取得
	 +
	 + 継承先でプレースホルダに設定する値を返すよう実装する
	 +/
	string[string] getPlaceHolderReplaces()
	{
		return null;
	}
}

/++
 + 必須入力をチェックする
 +/
class RequiredValidator : Validator
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +/
	this(Field field, string message)
	{
		super(field, message);
	}

	override
	bool validate()
	{
		return field_.value !is null && field_.value.length > 0;
	}
}

/++
 + パスワードの同一チェック
 +/
class PasswordConfirmValidator : Validator
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +  target = 同一内容を確認する対象 このフィールドと同じ内容ならばvalidateでtrueを返す
	 +/
	this(Field field, string message, Field target)
	{
		super(field, message);

		target_ = target;
	}

	override
	bool validate()
	{
		if (field_.value is null)
		{
			if (target_.value is null)
				return true;
		}
		else
		{
			if (target_.value !is null && std.string.cmp(field_.value, target_.value) == 0)
				return true;
		}

		return false;
	}

protected:

	/// 比較対象のフィールド
	Field target_;
}

/++
 + 入力値が正規表現に一致するか調べる
 +
 + 入力値が正規表現に一致した場合に、入力OKと判断する
 +/
class RegularExpressionValidator : Validator
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +  pattern = 正規表現 (ECMA standard)
	 +  attributes = 属性 (g, i, mの組み合わせ)
	 +/
	this(Field field, string message, string pattern, string attributes = null)
	{
		regex_ = regex(pattern, attributes);
		super(field, message);
	}

	override
	bool validate()
	{
		return !(match(field_.value, regex_).empty);
	}

protected:

	/// 使用する正規表現
	Regex!(char) regex_;
}

/++
 + 入力値が指定の範囲内の整数かチェック
 +
 + 入力値が32bit整数値の範囲外の場合不正とみなします
 +/
class IntegerRangeValidator : Validator
{
	/// 最小値に使うプレースホルダ
	static immutable
	string PLACE_HOLDER_MINIMUM = "_MINIMUM_";

	/// 最大値に使うプレースホルダ
	static immutable
	string PLACE_HOLDER_MAXIMUM = "_MAXIMUM_";

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +  min = 最小値
	 +  max = 最大値
	 +/
	this(Field field, string message, int min, int max)
	{
		// TODO: phobosのバグ対応で一回別の変数に代入する
		string phmin = PLACE_HOLDER_MINIMUM;
		string phmax = PLACE_HOLDER_MAXIMUM;
		message = std.array.replace(message, phmin, std.conv.to!string(min));
		message = std.array.replace(message, phmax, std.conv.to!string(max));

		super(field, message);

		min_ = min;
		max_ = max;
	}

	override
	bool validate()
	{
		if (field_.value is null || field_.value.length == 0)
			return false;

		int value;
		try
			value = std.conv.to!(int)(field_.value);
		catch (Error e)
			return false;

		return min_ <= value && value <= max_;
	}
	unittest
	{
		auto form = new Form("", "");
		auto field = new TextInputField(form, "test", "テスト");
		auto validator = new IntegerRangeValidator(field, "", 10, 99);

		field.value = "0";
		assert(!validator.validate());

		field.value = "10";
		assert(validator.validate());

		field.value = "50";
		assert(validator.validate());

		field.value = "99";
		assert(validator.validate());

		field.value = "100";
		assert(!validator.validate());
	}

protected:

	/// 最小値
	int min_;

	/// 最大値
	int max_;
}

/++
 + 文字列の長さをチェックするValidatorの基底クラス
 +
 + 似たような処理頻発したのでまとめてみただけ
 + 他のValidatorを作るときに継承することもないだろう
 +/
class LengthValidator : Validator
{
	/// 入力値の長さに使うプレースホルダ
	static immutable
	string PLACE_HOLDER_LENGTH = "_LENGTH_";

	/// 最小値に使うプレースホルダ
	static immutable
	string PLACE_HOLDER_MINIMUM = "_MINIMUM_";

	/// 最大値に使うプレースホルダ
	static immutable
	string PLACE_HOLDER_MAXIMUM = "_MAXIMUM_";

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +  threshold = 閾値
	 +  thresholdPlaceHolder = 閾値に置き換わるプレースホルダ名
	 +/
	this(Field field, string message, uint threshold, string thresholdPlaceHolder)
	{
		threshold_ = threshold;
		thresholdPlaceHolder_ = thresholdPlaceHolder;

		// ここで確定するので、制限の文字列長は埋め込んでおく
		message = std.array.replace(message, thresholdPlaceHolder_, std.conv.to!(string)(threshold_));

		string[] placeHolders = [PLACE_HOLDER_LENGTH];
		super(field, message, placeHolders);
	}

protected:

	/// 入力されたデータの長さ
	uint length_;

	/// 閾値
	uint threshold_;

	/// 閾値をメッセージに含めるためのプレースホルダ
	string thresholdPlaceHolder_;

	override
	string[string] getPlaceHolderReplaces()
	{
		string[string] replaces;
		replaces[PLACE_HOLDER_LENGTH] = to!(string)(length_);
		return replaces;
	}
}

/++
 + 最小文字列長以上かチェック
 +
 + コンストラクタに、許容する最小文字列長を指定してインスタンスを生成する
 +
 + validateメソッドでは、チェック対象のフィールドの値の文字列長が、
 + コンストラクタで指定した値より小さい場合にエラーとする
 +
 + Byteサイズではなく文字数であることに注意すること
 +
 + 使用可能なプレースホルダ
 + _LENGTH_ = 入力文字の長さ
 + _MINIMUM_ = 設定した最小文字列長
 +/
class MinimumLengthValidator : LengthValidator
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +  minimum = 最小文字列長
	 +/
	this(Field field, string message, uint minimum)
	{
		super(field, message, minimum, PLACE_HOLDER_MINIMUM);
	}

	override
	bool validate()
	{
		if (field_.value is null || field_.value.length == 0)
		{
			length_ = 0;
			return threshold_ == 0;
		}

		immutable(char)* iterator = field_.value.ptr;
		immutable(char)* end = iterator + field_.value.length;

		uint length = 0;
		while (iterator < end)
		{
			if ((*iterator & 0b01000000) == 0)
				++iterator;
			else if ((*iterator & 0b11100000) == 0b11000000)
				iterator += 2;
			else if ((*iterator & 0b11110000) == 0b11100000)
				iterator += 3;
			else if ((*iterator & 0b11111000) == 0b11110000)
				iterator += 4;
			else
				++iterator;

			++length;
		}

		length_ = length;

		return threshold_ <= length;
	}
}

/++
 + 最大文字列長以内かチェック
 +
 + コンストラクタに、許容する最大文字列長を指定してインスタンスを生成する
 +
 + validateメソッドでは、チェック対象のフィールドの値の文字列長が、
 + コンストラクタで指定した値より大きい場合にエラーとする
 +
 + Byteサイズではなく文字数であることに注意すること
 +
 + 使用可能なプレースホルダ
 + _LENGTH_ = 入力文字の長さ
 + _MAXIMUM_ = 設定した最大文字列長
 +/
class MaximumLengthValidator : LengthValidator
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +  maximum = 最大文字列長
	 +/
	this(Field field, string message, uint maximum)
	{
		super(field, message, maximum, PLACE_HOLDER_MAXIMUM);
	}

	override
	bool validate()
	{
		if (field_.value is null || field_.value.length == 0)
		{
			length_ = 0;
			return true;
		}

		immutable(char)* iterator = field_.value.ptr;
		immutable(char)* end = iterator + field_.value.length;

		uint length = 0;
		while (iterator < end)
		{
			if ((*iterator & 0b01000000) == 0)
				++iterator;
			else if ((*iterator & 0b11100000) == 0b11000000)
				iterator += 2;
			else if ((*iterator & 0b11110000) == 0b11100000)
				iterator += 3;
			else if ((*iterator & 0b11111000) == 0b11110000)
				iterator += 4;
			else
				++iterator;

			++length;
		}

		length_ = length;

		return threshold_ >= length;
	}
}

/++
 + 最小Byteサイズ以上かチェック
 +
 + コンストラクタに、許容する最小Byteサイズを指定してインスタンスを生成する
 +
 + validateメソッドでは、チェック対象のフィールドの値のByteサイズが、
 + コンストラクタで指定した値より小さい場合にエラーとする
 +
 + 文字数ではなくByteサイズであることに注意すること
 +
 + 使用可能なプレースホルダ
 + _LENGTH_ = 入力文字の長さ
 + _MINIMUM_ = 設定した最小Byteサイズ
 +/
class MinimumByteValidator : LengthValidator
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +  minimum = 最小文字列長
	 +/
	this(Field field, string message, uint minimum)
	{
		super(field, message, minimum, PLACE_HOLDER_MINIMUM);
	}

	override
	bool validate()
	{
		if (field_.value is null)
		{
			length_ = 0;
			return threshold_ == 0;
		}

		length_ = cast(uint)(field_.value.length);
		return threshold_ <= length_;
	}
}

/++
 + 最大Byteサイズ以内かチェック
 +
 + コンストラクタに、許容する最大Byteサイズを指定してインスタンスを生成する
 +
 + validateメソッドでは、チェック対象のフィールドの値のByteサイズが、
 + コンストラクタで指定した値より大きい場合にエラーとする
 +
 + 文字数ではなくByteサイズであることに注意すること
 +
 + 使用可能なプレースホルダ
 + _LENGTH_ = 入力文字の長さ
 + _MAXIMUM_ = 設定した最大Byteサイズ
 +/
class MaximumByteValidator : LengthValidator
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +  maximum = 最大バイト数
	 +/
	this(Field field, string message, uint maximum)
	{
		super(field, message, maximum, PLACE_HOLDER_MAXIMUM);
	}

	override
	bool validate()
	{
		if (field_.value is null)
		{
			length_ = 0;
			return true;
		}

		length_ = cast(uint)(field_.value.length);
		return threshold_ >= length_;
	}
}

/++
 + アップロードされたファイルが最小Byteサイズ以上かチェック
 +
 + コンストラクタに、許容する最小Byteサイズを指定してインスタンスを生成する
 +
 + validateメソッドでは、アップロードされたファイルのByteサイズが、
 + コンストラクタで指定した値より小さい場合にエラーとする
 +
 + ファイル名ではなく、ファイル自体のサイズをチェックするので注意すること
 +
 + 使用可能なプレースホルダ
 + _LENGTH_ = 入力文字の長さ
 + _MINIMUM_ = 設定した最小Byteサイズ
 +/
class UploadFileMinimumValidator : LengthValidator
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +  minimum = 最小文字列長
	 +/
	this(FileUploadField field, string message, uint minimum)
	{
		super(field, message, minimum, PLACE_HOLDER_MINIMUM);
	}

	override
	bool validate()
	{
		FileUploadField field = cast(FileUploadField)field_;
		if (field.uploadedFile_ is null)
		{
			length_ = 0;
			return threshold_ == 0;
		}

		length_ = field.uploadedFile_.size;
		return threshold_ <= length_;
	}
}

/++
 + アップロードされたファイルが最大Byteサイズ以下かチェック
 +
 + コンストラクタに、許容する最大Byteサイズを指定してインスタンスを生成する
 +
 + validateメソッドでは、アップロードされたファイルのByteサイズが、
 + コンストラクタで指定した値より大きい場合にエラーとする
 +
 + ファイル名ではなく、ファイル自体のサイズをチェックするので注意すること
 +
 + 使用可能なプレースホルダ
 + _LENGTH_ = 入力文字の長さ
 + _MINIMUM_ = 設定した最小Byteサイズ
 +/
class UploadFileMaximumValidator : LengthValidator
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  field = 対象の入力フィールド
	 +  message = 変換時のメッセージ
	 +  maximum = 最大バイト数
	 +/
	this(FileUploadField field, string message, uint maximum)
	{
		super(field, message, maximum, PLACE_HOLDER_MAXIMUM);
	}

	override
	bool validate()
	{
		FileUploadField field = cast(FileUploadField)field_;
		if (field.uploadedFile_ is null)
		{
			length_ = 0;
			return true;
		}

		length_ = field.uploadedFile_.size;
		return threshold_ >= length_;
	}
}
